import logging
import time
from functools import partial

import numpy as np
from scipy.sparse import csr_matrix

from . import kl_divergence
from .tsne import TSNEEmbedding

log = logging.getLogger(__name__)


class Callback:
    def optimzation_about_to_start(self):
        """This is called at the beginning of the optimization procedure."""

    def __call__(self, iteration: int, error: float, embedding: TSNEEmbedding) -> bool:
        """This is the main method called from the optimization."""


class ErrorLogger(Callback):
    def __init__(self):
        self.iter_count = 0
        self.last_log_time = None

    def optimzation_about_to_start(self):
        self.last_log_time = time.time()
        self.iter_count = 0

    def __call__(self, iteration, error, embedding):
        now = time.time()
        duration = now - self.last_log_time
        self.last_log_time = now

        n_iters = iteration - self.iter_count
        self.iter_count = iteration

        print('Iteration % 4d, KL divergence % 6.4f, %d iterations in %.4f sec' % (
            iteration, error, n_iters, duration))


class VerifyExaggerationError(Callback):
    """Used to verify that the exaggeration correction implemented in
    `gradient_descent` is correct."""
    def __init__(self, embedding: TSNEEmbedding) -> None:
        self.embedding = embedding
        # Keep a copy of the unexaggerated affinity matrix
        self.P = self.embedding.affinities.P.copy()

    def __call__(self, iteration: int, corrected_error: float, embedding: TSNEEmbedding):
        params = self.embedding.gradient_descent_params
        method = params['negative_gradient_method']

        if np.sum(embedding.affinities.P) <= 1:
            log.warning('Are you sure you are testing an exaggerated P matrix?')

        if method == 'fft':
            f = partial(kl_divergence.kl_divergence_approx_fft,
                        n_interpolation_points=params['n_interpolation_points'],
                        min_num_intervals=params['min_num_intervals'],
                        ints_in_interval=params['ints_in_interval'])
        elif method == 'bh':
            f = partial(kl_divergence.kl_divergence_approx_bh, theta=params['theta'])

        P = self.P

        true_error = f(P.indices, P.indptr, P.data, embedding)
        if abs(true_error - corrected_error) > 1e-8:
            raise RuntimeError('Correction term is wrong.')
        else:
            log.info('Corrected: %.4f - True %.4f [eps %.4f]' % (
                corrected_error, true_error, abs(true_error - corrected_error)))


class ErrorApproximations(Callback):
    """Check how good the error approximations are. Of course, we use an
    approximation for P so this itself is an approximation."""
    def __init__(self, P: csr_matrix):
        self.P = P.copy()
        self.exact_errors = []
        self.bh_errors = []
        self.fft_errors = []

    def __call__(self, iteration: int, error: float, embedding: TSNEEmbedding):
        exact_error = kl_divergence.kl_divergence_exact(self.P.toarray(), embedding)
        bh_error = kl_divergence.kl_divergence_approx_bh(
            self.P.indices, self.P.indptr, self.P.data, embedding)
        fft_error = kl_divergence.kl_divergence_approx_fft(
            self.P.indices, self.P.indptr, self.P.data, embedding)

        self.exact_errors.append(exact_error)
        self.bh_errors.append(bh_error)
        self.fft_errors.append(fft_error)

    def report(self):
        exact_errors = np.array(self.exact_errors)
        bh_errors = np.array(self.bh_errors)
        fft_errors = np.array(self.fft_errors)

        bh_diff = bh_errors - exact_errors
        print('Barnes-Hut: mean difference %.4f (±%.4f)' % (
            np.mean(bh_diff), np.std(bh_diff)))

        fft_diff = fft_errors - exact_errors
        print('Interpolation: mean difference %.4f (±%.4f)' % (
            np.mean(fft_diff), np.std(fft_diff)))



# Author: Leland McInnes <leland.mcinnes@gmail.com>
#
# License: BSD 2 clause
import numba
import numpy as np

from collections import namedtuple

from fastTSNE.pynndescent.utils import tau_rand_int, norm


RandomProjectionTreeNode = namedtuple('RandomProjectionTreeNode',
                                      ['indices', 'is_leaf', 'hyperplane',
                                       'offset', 'left_child', 'right_child'])

FlatTree = namedtuple('FlatTree', ['hyperplanes', 'offsets',
                                   'children', 'indices'])


@numba.njit(fastmath=True, nogil=True, parallel=True)
def euclidean_random_projection_split(data, indices, rng_state):
    """Given a set of ``indices`` for data points from ``data``, create
    a random hyperplane to split the data, returning two arrays indices
    that fall on either side of the hyperplane. This is the basis for a
    random projection tree, which simply uses this splitting recursively.

    This particular split uses euclidean distance to determine the hyperplane
    and which side each data sample falls on.

    Parameters
    ----------
    data: array of shape (n_samples, n_features)
        The original data to be split

    indices: array of shape (tree_node_size,)
        The indices of the elements in the ``data`` array that are to
        be split in the current operation.

    rng_state: array of int64, shape (3,)
        The internal state of the rng

    Returns
    -------
    indices_left: array
        The elements of ``indices`` that fall on the "left" side of the
        random hyperplane.

    indices_right: array
        The elements of ``indices`` that fall on the "left" side of the
        random hyperplane.
    """
    dim = data.shape[1]

    # Select two random points, set the hyperplane between them
    left_index = tau_rand_int(rng_state) % indices.shape[0]
    right_index = tau_rand_int(rng_state) % indices.shape[0]
    right_index += left_index == right_index
    right_index = right_index % indices.shape[0]
    left = indices[left_index]
    right = indices[right_index]

    # Compute the normal vector to the hyperplane (the vector between
    # the two points) and the offset from the origin
    hyperplane_offset = 0.0
    hyperplane_vector = np.empty(dim, dtype=np.float32)

    for d in range(dim):
        hyperplane_vector[d] = data[left, d] - data[right, d]
        hyperplane_offset -= hyperplane_vector[d] * (
            data[left, d] + data[right, d]) / 2.0

    # For each point compute the margin (project into normal vector, add offset)
    # If we are on lower side of the hyperplane put in one pile, otherwise
    # put it in the other pile (if we hit hyperplane on the nose, flip a coin)
    n_left = 0
    n_right = 0
    side = np.empty(indices.shape[0], np.int8)
    for i in range(indices.shape[0]):
        margin = hyperplane_offset
        for d in range(dim):
            margin += hyperplane_vector[d] * data[indices[i], d]

        if margin == 0:
            side[i] = tau_rand_int(rng_state) % 2
            if side[i] == 0:
                n_left += 1
            else:
                n_right += 1
        elif margin > 0:
            side[i] = 0
            n_left += 1
        else:
            side[i] = 1
            n_right += 1

    # Now that we have the counts allocate arrays
    indices_left = np.empty(n_left, dtype=np.int64)
    indices_right = np.empty(n_right, dtype=np.int64)

    # Populate the arrays with indices according to which side they fell on
    n_left = 0
    n_right = 0
    for i in range(side.shape[0]):
        if side[i] == 0:
            indices_left[n_left] = indices[i]
            n_left += 1
        else:
            indices_right[n_right] = indices[i]
            n_right += 1

    return indices_left, indices_right, hyperplane_vector, hyperplane_offset


@numba.njit()
def angular_random_projection_split(data, indices, rng_state):
    """Given a set of ``indices`` for data points from ``data``, create
    a random hyperplane to split the data, returning two arrays indices
    that fall on either side of the hyperplane. This is the basis for a
    random projection tree, which simply uses this splitting recursively.

    This particular split uses cosine distance to determine the hyperplane
    and which side each data sample falls on.

    Parameters
    ----------
    data: array of shape (n_samples, n_features)
        The original data to be split

    indices: array of shape (tree_node_size,)
        The indices of the elements in the ``data`` array that are to
        be split in the current operation.

    rng_state: array of int64, shape (3,)
        The internal state of the rng

    Returns
    -------
    indices_left: array
        The elements of ``indices`` that fall on the "left" side of the
        random hyperplane.

    indices_right: array
        The elements of ``indices`` that fall on the "left" side of the
        random hyperplane.
    """
    dim = data.shape[1]

    # Select two random points, set the hyperplane between them
    left_index = tau_rand_int(rng_state) % indices.shape[0]
    right_index = tau_rand_int(rng_state) % indices.shape[0]
    right_index += left_index == right_index
    right_index = right_index % indices.shape[0]
    left = indices[left_index]
    right = indices[right_index]

    left_norm = norm(data[left])
    right_norm = norm(data[right])

    if left_norm == 0.0:
        left_norm = 1.0

    if right_norm == 0.0:
        right_norm = 1.0

    # Compute the normal vector to the hyperplane (the vector between
    # the two points)
    hyperplane_vector = np.empty(dim, dtype=np.float32)

    for d in range(dim):
        hyperplane_vector[d] = ((data[left, d] / left_norm) -
                                (data[right, d] / right_norm))

    hyperplane_norm = norm(hyperplane_vector)
    if hyperplane_norm == 0.0:
        hyperplane_norm = 1.0

    for d in range(dim):
        hyperplane_vector[d] = hyperplane_vector[d] / hyperplane_norm

    # For each point compute the margin (project into normal vector)
    # If we are on lower side of the hyperplane put in one pile, otherwise
    # put it in the other pile (if we hit hyperplane on the nose, flip a coin)
    n_left = 0
    n_right = 0
    side = np.empty(indices.shape[0], np.int8)
    for i in range(indices.shape[0]):
        margin = 0.0
        for d in range(dim):
            margin += hyperplane_vector[d] * data[indices[i], d]

        if margin == 0:
            side[i] = tau_rand_int(rng_state) % 2
            if side[i] == 0:
                n_left += 1
            else:
                n_right += 1
        elif margin > 0:
            side[i] = 0
            n_left += 1
        else:
            side[i] = 1
            n_right += 1

    # Now that we have the counts allocate arrays
    indices_left = np.empty(n_left, dtype=np.int64)
    indices_right = np.empty(n_right, dtype=np.int64)

    # Populate the arrays with indices according to which side they fell on
    n_left = 0
    n_right = 0
    for i in range(side.shape[0]):
        if side[i] == 0:
            indices_left[n_left] = indices[i]
            n_left += 1
        else:
            indices_right[n_right] = indices[i]
            n_right += 1

    return indices_left, indices_right, hyperplane_vector, 0


def make_euclidean_tree(data, indices, rng_state, leaf_size=30):
    if indices.shape[0] > leaf_size:
        left_indices, right_indices, hyperplane, offset = \
            euclidean_random_projection_split(data, indices, rng_state)

        left_node = make_euclidean_tree(data,
                                        left_indices,
                                        rng_state,
                                        leaf_size)
        right_node = make_euclidean_tree(data,
                                         right_indices,
                                         rng_state,
                                         leaf_size)

        node = RandomProjectionTreeNode(None, False, hyperplane, offset,
                                        left_node, right_node)
    else:
        node = RandomProjectionTreeNode(indices, True, None, None, None, None)

    return node


def make_angular_tree(data, indices, rng_state, leaf_size=30):
    if indices.shape[0] > leaf_size:
        left_indices, right_indices, hyperplane, offset = \
            angular_random_projection_split(data, indices, rng_state)

        left_node = make_angular_tree(data,
                                      left_indices,
                                      rng_state,
                                      leaf_size)
        right_node = make_angular_tree(data,
                                       right_indices,
                                       rng_state,
                                       leaf_size)

        node = RandomProjectionTreeNode(None, False, hyperplane, offset,
                                        left_node, right_node)
    else:
        node = RandomProjectionTreeNode(indices, True, None, None, None, None)

    return node


def num_nodes(tree):
    """Determine the number of nodes in a tree"""
    if tree.is_leaf:
        return 1
    else:
        return 1 + num_nodes(tree.left_child) + num_nodes(tree.right_child)


def num_leaves(tree):
    """Determine the number of leaves in a tree"""
    if tree.is_leaf:
        return 1
    else:
        return num_leaves(tree.left_child) + num_leaves(tree.right_child)


def recursive_flatten(tree, hyperplanes, offsets,
                      children, indices, node_num,
                      leaf_num):
    if tree.is_leaf:
        children[node_num, 0] = -leaf_num
        indices[leaf_num, :tree.indices.shape[0]] = tree.indices
        leaf_num += 1
        return node_num, leaf_num
    else:
        hyperplanes[node_num] = tree.hyperplane
        offsets[node_num] = tree.offset
        children[node_num, 0] = node_num + 1
        old_node_num = node_num
        node_num, leaf_num = recursive_flatten(tree.left_child,
                                               hyperplanes, offsets,
                                               children, indices,
                                               node_num + 1, leaf_num)
        children[old_node_num, 1] = node_num + 1
        node_num, leaf_num = recursive_flatten(tree.right_child,
                                               hyperplanes, offsets,
                                               children, indices,
                                               node_num + 1, leaf_num)
        return node_num, leaf_num


def flatten_tree(tree, leaf_size):
    n_nodes = num_nodes(tree)
    n_leaves = num_leaves(tree)
    hyperplanes = np.zeros((n_nodes, tree.hyperplane.shape[0]),
                           dtype=np.float32)
    offsets = np.zeros(n_nodes, dtype=np.float32)
    children = -1 * np.ones((n_nodes, 2), dtype=np.int64)
    indices = -1 * np.ones((n_leaves, leaf_size), dtype=np.int64)
    recursive_flatten(tree, hyperplanes, offsets, children, indices, 0, 0)
    return FlatTree(hyperplanes, offsets, children, indices)


@numba.njit(fastmath=True)
def select_side(hyperplane, offset, point, rng_state):
    margin = offset
    for d in range(point.shape[0]):
        margin += hyperplane[d] * point[d]

    if margin == 0:
        side = tau_rand_int(rng_state) % 2
        if side == 0:
            return 0
        else:
            return 1
    elif margin > 0:
        return 0
    else:
        return 1


@numba.njit(fastmath=True)
def search_flat_tree(point,
                     hyperplanes, offsets,
                     children, indices,
                     rng_state):
    node = 0
    while children[node, 0] > 0:
        side = select_side(hyperplanes[node], offsets[node], point, rng_state)
        if side == 0:
            node = children[node, 0]
        else:
            node = children[node, 1]

    return indices[-children[node, 0]]



import logging

import numpy as np
from scipy.sparse import csr_matrix

from . import _tsne
from .nearest_neighbors import BallTree, NNDescent, KNNIndex, VALID_METRICS

log = logging.getLogger(__name__)


class Affinities:
    """Compute the affinities among some initial data and new data.

    t-SNE takes as input an affinity matrix :math:`P`, and does not really care
    about anything else about the data. This means we can use t-SNE for any data
    where we are able to express interactions between samples with an affinity
    matrix.

    Attributes
    ----------
    P: array_like
        The affinity matrix expressing interactions between all data samples.

    """

    def __init__(self):
        self.P = None

    def to_new(self, data):
        """Compute the affinities of new data points to the existing ones.

        This is especially useful for `transform` where we need the conditional
        probabilities from the existing to the new data.

        """


class PerplexityBasedNN(Affinities):
    """Compute affinities using the nearest neighbors defined by perplexity.

    Parameters
    ----------
    data: np.ndarray
        The data matrix.
    perplexity: float
        Perplexity can be thought of as the continuous :math:`k` number of
        neighbors to consider for each data point. To avoid confusion, note that
        perplexity linearly impacts runtime.
    method: str
        Specifies the nearest neighbor method to use. Can be either ``exact`` or
        ``approx``. ``exact`` uses space partitioning binary trees from
        scikit-learn while ``approx`` makes use of nearest neighbor descent.
        Note that ``approx`` has a bit of overhead and will be slower on smaller
        data sets than exact search.
    metric: str
        The metric to be used to compute affinities between points in the
        original space.
    metric_params: Optional[dict]
        Additional keyword arguments for the metric function.
    symmetrize: bool
        Symmetrize affinity matrix. Standard t-SNE symmetrizes the interactions
        but when embedding new data, symmetrization is not performed.
    n_jobs: int
        The number of jobs to run in parallel. This follows the scikit-learn
        convention, ``-1`` meaning all processors, ``-2`` meaning all but one
        processor and so on.
    random_state: Optional[Union[int, RandomState]]
        The random state parameter follows the convention used in scikit-learn.
        If the value is an int, random_state is the seed used by the random
        number generator. If the value is a RandomState instance, then it will
        be used as the random number generator. If the value is None, the random
        number generator is the RandomState instance used by `np.random`.

    """

    def __init__(self, data, perplexity=30, method='exact', metric='euclidean',
                 metric_params=None, symmetrize=True, n_jobs=1, random_state=None):
        self.n_samples = data.shape[0]
        self.perplexity = self.check_perplexity(perplexity)

        self.knn_index = build_knn_index(data, method, metric, metric_params, n_jobs, random_state)

        # Find and store the nearest neighbors so we can reuse them if the
        # perplexity is ever lowered
        k_neighbors = min(self.n_samples - 1, int(3 * self.perplexity))
        self.__neighbors, self.__distances = self.knn_index.query_train(data, k=k_neighbors)

        self.P = joint_probabilities_nn(self.__neighbors, self.__distances, [self.perplexity],
                                        symmetrize=symmetrize, n_jobs=n_jobs)

        self.n_jobs = n_jobs

    def set_perplexity(self, new_perplexity):
        # If the value hasn't changed, there's nothing to do
        if new_perplexity == self.perplexity:
            return
        # Verify that the perplexity isn't too large
        new_perplexity = self.check_perplexity(new_perplexity)
        # Recompute the affinity matrix
        k_neighbors = min(self.n_samples - 1, int(3 * new_perplexity))
        if k_neighbors > self.__neighbors.shape[1]:
            raise RuntimeError(
                'The desired perplexity `%.2f` is larger than the initial one '
                'used. This would need to recompute the nearest neighbors, '
                'which is not efficient. Please create a new `%s` instance '
                'with the increased perplexity.' % (
                    new_perplexity, self.__class__.__name__))

        self.perplexity = new_perplexity
        self.P = joint_probabilities_nn(
            self.__neighbors[:, :k_neighbors], self.__distances[:, :k_neighbors],
            [self.perplexity], symmetrize=True, n_jobs=self.n_jobs,
        )

    def to_new(self, data, perplexity=None, return_distances=False):
        perplexity = perplexity if perplexity is not None else self.perplexity
        perplexity = self.check_perplexity(perplexity)
        k_neighbors = min(self.n_samples - 1, int(3 * perplexity))

        neighbors, distances = self.knn_index.query(data, k_neighbors)

        P = joint_probabilities_nn(
            neighbors, distances, [perplexity], symmetrize=False,
            n_reference_samples=self.n_samples, n_jobs=self.n_jobs,
        )

        if return_distances:
            return P, neighbors, distances

        return P

    def check_perplexity(self, perplexity):
        """Check for valid perplexity value."""
        if perplexity <= 0:
            raise ValueError('Perplexity must be >=0. %.2f given' % perplexity)

        if self.n_samples - 1 < 3 * perplexity:
            old_perplexity, perplexity = perplexity, (self.n_samples - 1) / 3
            log.warning('Perplexity value %d is too high. Using perplexity '
                        '%.2f instead' % (old_perplexity, perplexity))

        return perplexity


def build_knn_index(data, method, metric, metric_params=None, n_jobs=1, random_state=None):
    methods = {'exact': BallTree, 'approx': NNDescent}
    if isinstance(method, KNNIndex):
        knn_index = method

    elif method not in methods:
        raise ValueError('Unrecognized nearest neighbor algorithm `%s`. '
                         'Please choose one of the supported methods or '
                         'provide a valid `KNNIndex` instance.' % method)
    else:
        if metric not in VALID_METRICS:
            raise ValueError('Unrecognized distance metric `%s`. Please '
                             'choose one of the supported methods.' % metric)
        knn_index = methods[method](metric=metric, metric_params=metric_params,
                                    n_jobs=n_jobs, random_state=random_state)

    knn_index.build(data)

    return knn_index


def joint_probabilities_nn(neighbors, distances, perplexities, symmetrize=True,
                           n_reference_samples=None, n_jobs=1):
    """Compute the conditional probability matrix P_{j|i}.

    This method computes an approximation to P using the nearest neighbors.

    Parameters
    ----------
    neighbors: np.ndarray
        A `n_samples * k_neighbors` matrix containing the indices to each
        points' nearest neighbors in descending order.
    distances: np.ndarray
        A `n_samples * k_neighbors` matrix containing the distances to the
        neighbors at indices defined in the neighbors parameter.
    perplexities: double
        The desired perplexity of the probability distribution.
    symmetrize: bool
        Whether to symmetrize the probability matrix or not. Symmetrizing is
        used for typical t-SNE, but does not make sense when embedding new data
        into an existing embedding.
    n_reference_samples: int
        The number of samples in the existing (reference) embedding. Needed to
        properly construct the sparse P matrix.
    n_jobs: int
        Number of threads.

    Returns
    -------
    csr_matrix
        A `n_samples * n_reference_samples` matrix containing the probabilities
        that a new sample would appear as a neighbor of a reference point.

    """
    n_samples, k_neighbors = distances.shape

    if n_reference_samples is None:
        n_reference_samples = n_samples

    # Compute asymmetric pairwise input similarities
    conditional_P = _tsne.compute_gaussian_perplexity(
        distances, np.array(perplexities, dtype=float), num_threads=n_jobs)
    conditional_P = np.asarray(conditional_P)

    P = csr_matrix((conditional_P.ravel(), neighbors.ravel(),
                    range(0, n_samples * k_neighbors + 1, k_neighbors)),
                   shape=(n_samples, n_reference_samples))

    # Symmetrize the probability matrix
    if symmetrize:
        P = (P + P.T) / 2

    # Convert weights to probabilities
    P /= np.sum(P)

    return P


class FixedSigmaNN(Affinities):

    def __init__(self, data, sigma, k=30, method='exact', metric='euclidean',
                 metric_params=None, symmetrize=True, n_jobs=1, random_state=None):
        self.n_samples = n_samples = data.shape[0]

        if k >= self.n_samples:
            raise ValueError('`k` (%d) cannot be larger than N-1 (%d).' % (k, self.n_samples))

        knn_index = build_knn_index(data, method, metric, metric_params, n_jobs, random_state)
        neighbors, distances = knn_index.query_train(data, k=k)

        self.knn_index = knn_index

        # Compute asymmetric pairwise input similarities
        conditional_P = np.exp(-distances ** 2 / (2 * sigma ** 2))
        conditional_P /= np.sum(conditional_P, axis=1)[:, np.newaxis]

        P = csr_matrix((conditional_P.ravel(), neighbors.ravel(),
                        range(0, n_samples * k + 1, k)),
                       shape=(n_samples, n_samples))

        # Symmetrize the probability matrix
        if symmetrize:
            P = (P + P.T) / 2

        # Convert weights to probabilities
        P /= np.sum(conditional_P)

        self.sigma = sigma
        self.k = k
        self.P = P
        self.n_jobs = n_jobs

    def to_new(self, data, k=None, sigma=None, return_distances=False):
        n_samples = data.shape[0]
        n_reference_samples = self.n_samples

        if k is None:
            k = self.k
        elif k >= n_reference_samples:
            raise ValueError('`k` (%d) cannot be larger than the number of '
                             'reference samples (%d).' % (k, self.n_samples))

        if sigma is None:
            sigma = self.sigma

        # Find nearest neighbors and the distances to the new points
        neighbors, distances = self.knn_index.query(data, k)

        # Compute asymmetric pairwise input similarities
        conditional_P = np.exp(-distances ** 2 / (2 * sigma ** 2))
        conditional_P /= np.sum(conditional_P, axis=1)[:, np.newaxis]

        P = csr_matrix((conditional_P.ravel(), neighbors.ravel(),
                        range(0, n_samples * k + 1, k)),
                       shape=(n_samples, n_reference_samples))

        # Convert weights to probabilities
        P /= np.sum(conditional_P)

        if return_distances:
            return P, neighbors, distances

        return P


class Multiscale(Affinities):
    def __init__(self, data, perplexities, method='exact', metric='euclidean',
                 metric_params=None, symmetrize=True, n_jobs=1, random_state=None):
        self.n_samples = data.shape[0]

        # We will compute the nearest neighbors to the max value of perplexity,
        # smaller values can just use indexing to truncate unneeded neighbors
        perplexities = self.check_perplexities(perplexities)
        max_perplexity = np.max(perplexities)
        k_neighbors = min(self.n_samples - 1, int(3 * max_perplexity))

        knn_index = build_knn_index(data, method, metric, metric_params, n_jobs, random_state)
        neighbors, distances = knn_index.query_train(data, k=k_neighbors)

        self.knn_index = knn_index
        self.P = joint_probabilities_nn(neighbors, distances, perplexities,
                                        symmetrize=symmetrize, n_jobs=n_jobs)

        self.perplexities = perplexities
        self.n_jobs = n_jobs

    def to_new(self, data, perplexities=None, return_distances=False):
        perplexities = perplexities if perplexities is not None else self.perplexities
        perplexities = self.check_perplexities(perplexities)

        max_perplexity = np.max(perplexities)
        k_neighbors = min(self.n_samples - 1, int(3 * max_perplexity))

        neighbors, distances = self.knn_index.query(data, k_neighbors)

        P = joint_probabilities_nn(
            neighbors, distances, perplexities, symmetrize=False,
            n_reference_samples=self.n_samples, n_jobs=self.n_jobs,
        )

        if return_distances:
            return P, neighbors, distances

        return P

    def check_perplexities(self, perplexities):
        """Check and correct/truncate perplexities.

        If a perplexity is too large, it is corrected to the largest allowed
        value. It is then inserted into the list of perplexities only if that
        value doesn't already exist in the list.

        """
        usable_perplexities = []
        for perplexity in sorted(perplexities):
            if 3 * perplexity > self.n_samples - 1:
                new_perplexity = (self.n_samples - 1) / 3

                if new_perplexity in usable_perplexities:
                    log.warning('Perplexity value %d is too high. Dropping '
                                'because the max perplexity is already in the '
                                'list.' % perplexity)
                else:
                    usable_perplexities.append(new_perplexity)
                    log.warning('Perplexity value %d is too high. Using '
                                'perplexity %.2f instaed' % (perplexity, new_perplexity))
            else:
                usable_perplexities.append(perplexity)

        return usable_perplexities



import unittest

import numpy as np
from sklearn import datasets
from sklearn.metrics import accuracy_score
from sklearn.neighbors import KNeighborsClassifier

from fastTSNE.callbacks import VerifyExaggerationError
from fastTSNE.tsne import TSNE, TSNEEmbedding


class TestTSNECorrectness(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tsne = TSNE(early_exaggeration_iter=20, n_iter=100)
        # Set up two modalities, if we want to viually inspect test results
        random_state = np.random.RandomState(0)
        cls.x = np.vstack((
            random_state.normal(+1, 1, (100, 4)),
            random_state.normal(-1, 1, (100, 4)),
        ))
        cls.x_test = random_state.normal(0, 1, (25, 4))

    def test_basic_flow(self):
        """Verify that the basic flow does not crash."""
        embedding = self.tsne.fit(self.x)
        self.assertFalse(np.any(np.isnan(embedding)))

        partial_embedding = embedding.transform(
            self.x_test, early_exaggeration_iter=20, n_iter=20)
        self.assertFalse(np.any(np.isnan(partial_embedding)))

    def test_advanced_flow(self):
        """Verify that the advanced flow does not crash."""
        embedding = self.tsne.prepare_initial(self.x)
        embedding = embedding.optimize(20, exaggeration=12)
        embedding = embedding.optimize(20)  # type: TSNEEmbedding
        self.assertFalse(np.any(np.isnan(embedding)))

        partial_embedding = embedding.prepare_partial(self.x_test)
        partial_embedding = partial_embedding.optimize(20, exaggeration=2)
        partial_embedding = partial_embedding.optimize(20)
        self.assertFalse(np.any(np.isnan(partial_embedding)))

    def test_error_exaggeration_correction(self):
        embedding = self.tsne.prepare_initial(self.x)

        # The callback raises if the KL divergence does not match the true one
        embedding.optimize(
            50, exaggeration=5, callbacks=[VerifyExaggerationError(embedding)],
            callbacks_every_iters=1, inplace=True,
        )

    def test_iris(self):
        iris = datasets.load_iris()
        x, y = iris['data'], iris['target']

        # Evaluate tSNE optimization using a KNN classifier
        knn = KNeighborsClassifier(n_neighbors=10)
        tsne = TSNE(perplexity=30, initialization='random', random_state=0)

        # Prepare a random initialization
        embedding = tsne.prepare_initial(x)

        # KNN should do poorly on a random initialization
        knn.fit(embedding, y)
        predictions = knn.predict(embedding)
        self.assertTrue(accuracy_score(predictions, y) < .5)

        # Optimize the embedding for a small number of steps so tests run fast
        embedding.optimize(50, inplace=True)

        # Similar points should be grouped together, therefore KNN should do well
        knn.fit(embedding, y)
        predictions = knn.predict(embedding)
        self.assertTrue(accuracy_score(predictions, y) > .95)



import gzip
import os
import pickle
import urllib
from os.path import abspath, dirname, join

import numpy as np

FILE_DIR = dirname(abspath(__file__))
DATA_DIR = join(FILE_DIR, 'data')


def get_zeisel_2018(n_samples: int = None):
    with gzip.open(join(DATA_DIR, 'zeisel_2018.pkl.gz'), 'rb') as f:
        data = pickle.load(f)

    # Extract log normalized counts and cell type
    x, y = data['log_counts'], data['CellType1']
    x = x.T

    if n_samples is not None:
        indices = np.random.choice(list(range(x.shape[0])), n_samples, replace=False)
        x, y = x[indices], y[indices]

    return x, y


def get_mnist(n_samples: int = None):
    if not os.path.exists(join(DATA_DIR, 'mnist.pkl.gz')):
        urllib.request.urlretrieve(
            'http://deeplearning.net/data/mnist/mnist.pkl.gz', join(DATA_DIR, 'mnist.pkl.gz'))

    with gzip.open(join(DATA_DIR, 'mnist.pkl.gz'), 'rb') as f:
        train, val, test = pickle.load(f, encoding='latin1')
    _train = np.asarray(train[0], dtype=np.float64)
    _val = np.asarray(val[0], dtype=np.float64)
    _test = np.asarray(test[0], dtype=np.float64)
    x = np.vstack((_train, _val, _test))
    y = np.hstack((train[1], val[1], test[1]))

    if n_samples is not None:
        indices = np.random.choice(list(range(x.shape[0])), n_samples, replace=False)
        x, y = x[indices], y[indices]

    return x, y


def plot(x: np.ndarray, y: np.ndarray, ax=None, draw_legend=True, **kwargs) -> None:
    if ax is None:
        import matplotlib.pyplot as plt
        _, ax = plt.subplots()

    for yi in np.unique(y):
        mask = y == yi
        ax.plot(x[mask, 0], x[mask, 1], 'o', label=str(yi),
                alpha=kwargs.get('alpha', 0.6), ms=kwargs.get('ms', 1))
    
    # Hide ticks
    ax.set_xticks([]), ax.set_yticks([])
    # Hide box border around figure
    ax.axis('off')

    if draw_legend:
        legend = ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        # Make the legend markers bigger, so we can see what's what
        for lh in legend.legendHandles:
            lh._legmarker.set_markersize(6)
            lh._legmarker.set_alpha(1)



import numpy as np

from .tsne import TSNEEmbedding


def pBIC(embedding: TSNEEmbedding) -> float:
    if not hasattr(embedding.affinities, 'perplexity'):
        raise TypeError('The embedding affinity matrix has no attribute `perplexity`')
    n_samples = embedding.shape[0]

    return 2 * embedding.kl_divergence + np.log(n_samples) * \
        embedding.affinities.perplexity / n_samples



import inspect
import unittest
from functools import wraps
from typing import Callable, Any, Tuple, Optional
from unittest.mock import patch, MagicMock

import numpy as np

from fastTSNE.affinity import PerplexityBasedNN
from fastTSNE.nearest_neighbors import VALID_METRICS
from fastTSNE.tsne import TSNE, kl_divergence_bh, kl_divergence_fft
from fastTSNE import tsne

np.random.seed(42)


def check_params(params: dict) -> Callable:
    """Run a series of parameterized tests to check tSNE parameter flow."""

    def _decorator(test_case: Callable) -> Callable:
        @wraps(test_case)
        def _wrapper(self):
            for param_name in params:
                for param_value in params[param_name]:
                    test_case(self, param_name, param_value)

        return _wrapper

    return _decorator


def check_call_contains_kwargs(
        call: Tuple,
        params: dict,
        param_mapping: Optional[dict] = None,
) -> None:
    """Check whether a `call` object was called with some params, but also some
    others we don't care about"""
    _param_mapping = {'negative_gradient_method': 'objective_function',
                      'early_exaggeration_iter': 'n_iter',
                      'late_exaggeration_iter': 'n_iter',
                      'early_exaggeration': 'exaggeration',
                      'late_exaggeration': 'exaggeration',
                      'initial_momentum': 'momentum',
                      'final_momentum': 'momentum'}
    if param_mapping is not None:
        _param_mapping.update(param_mapping)

    name, args, kwargs = call
    for key in params:
        # If a parameter isn't named the same way in the call
        if key in _param_mapping:
            kwargs_key = _param_mapping[key]
        else:
            kwargs_key = key

        expected_value = params[key]
        actual_value = kwargs.get(kwargs_key, None)
        if expected_value != actual_value:
            raise AssertionError(
                'Mock not called with `%s=%s`. Called with `%s`' %
                (key, expected_value, actual_value)
            )


def check_mock_called_with_kwargs(mock: MagicMock, params: dict) -> None:
    """Check whether a mock was called with kwargs, but also some other params
    we don't care about."""
    for call in mock.mock_calls:
        check_call_contains_kwargs(call, params)


class TestTSNEParameterFlow(unittest.TestCase):
    """Test that the optimization parameters get properly propagated."""

    grad_descent_params = {
        'negative_gradient_method': [kl_divergence_bh, kl_divergence_fft],
        'learning_rate': [1, 10, 100],
        'theta': [0.2, 0.5, 0.8],
        'n_interpolation_points': [3, 5],
        'min_num_intervals': [10, 20, 30],
        'ints_in_interval': [1, 2, 5],
        'n_jobs': [1, 2, 4],
        'callbacks': [None, [lambda *args, **kwargs: ...]],
        'callbacks_every_iters': [25, 50],
    }

    @classmethod
    def setUpClass(cls):
        cls.x = np.random.randn(100, 4)
        cls.x_test = np.random.randn(25, 4)

    @check_params({**grad_descent_params, **{
        'early_exaggeration_iter': [50, 100],
        'early_exaggeration': [4, 12],
        'initial_momentum': [0.2, 0.5, 0.8],
        'n_iter': [50, 100],
        'exaggeration': [None, 2],
        'final_momentum': [0.2, 0.5, 0.8],
    }})
    @patch('fastTSNE.tsne.gradient_descent.__call__')
    def test_constructor(self, param_name, param_value, gradient_descent):
        # type: (str, Any, MagicMock) -> None
        # Make sure mock still conforms to signature
        gradient_descent.return_value = (1, MagicMock())

        # Early exaggeration training loop
        if param_name in ('early_exaggeration_iter', 'early_exaggeration', 'initial_momentum'):
            call_idx = 0
        # Main training loop
        elif param_name in ('n_iter', 'exaggeration', 'final_momentum'):
            call_idx = 1
        # If general parameter, should be applied to every call
        else:
            call_idx = 0

        TSNE(**{param_name: param_value}).fit(self.x)

        self.assertEqual(2, gradient_descent.call_count)
        check_call_contains_kwargs(
            gradient_descent.mock_calls[call_idx],
            {param_name: param_value},
        )

    @check_params({**grad_descent_params, **{
        'n_iter': [50, 100, 150],
        'exaggeration': [None, 2, 5],
        'momentum': [0.2, 0.5, 0.8],
    }})
    @patch('fastTSNE.tsne.gradient_descent.__call__')
    def test_embedding_optimize(self, param_name, param_value, gradient_descent):
        # type: (str, Any, MagicMock) -> None
        # Make sure mock still conforms to signature
        gradient_descent.return_value = (1, MagicMock())

        # `optimize` requires us to specify the `n_iter`
        params = {'n_iter': 50, param_name: param_value}

        tsne = TSNE()
        embedding = tsne.prepare_initial(self.x)
        embedding.optimize(**params, inplace=True)

        self.assertEqual(1, gradient_descent.call_count)
        check_call_contains_kwargs(gradient_descent.mock_calls[0], params)

    @check_params({
        'early_exaggeration_iter': [50, 100],
        'early_exaggeration': [4, 12],
        'initial_momentum': [0.2, 0.5, 0.8],
        'n_iter': [50, 100],
        'final_momentum': [0.2, 0.5, 0.8],
    })
    @patch('fastTSNE.tsne.gradient_descent.__call__')
    def test_embedding_transform(self, param_name, param_value, gradient_descent):
        # type: (str, Any, MagicMock) -> None
        # Make sure mock still conforms to signature
        gradient_descent.return_value = (1, MagicMock())

        # Perform initial embedding - this is tested above
        tsne = TSNE()
        embedding = tsne.fit(self.x)
        gradient_descent.reset_mock()

        embedding.transform(self.x_test, **{param_name: param_value})

        # Early exaggeration training loop
        if param_name in ('early_exaggeration_iter', 'early_exaggeration'):
            call_idx = 0
        # Main training loop
        elif param_name in ('n_iter', 'final_momentum'):
            call_idx = 1

        # If general parameter, should be applied to every call
        else:
            call_idx = 0

        self.assertEqual(2, gradient_descent.call_count)
        check_call_contains_kwargs(
            gradient_descent.mock_calls[call_idx],
            {param_name: param_value},
        )

    @check_params({**grad_descent_params, **{
        'n_iter': [50, 100, 150],
        'exaggeration': [None, 2, 5],
        'momentum': [0.2, 0.5, 0.8],
    }})
    @patch('fastTSNE.tsne.gradient_descent.__call__')
    def test_partial_embedding_optimize(self, param_name, param_value, gradient_descent):
        # type: (str, Any, MagicMock) -> None
        # Make sure mock still conforms to signature
        gradient_descent.return_value = (1, MagicMock())

        # Perform initial embedding - this is tested above
        tsne = TSNE()
        embedding = tsne.fit(self.x)
        gradient_descent.reset_mock()

        # `optimize` requires us to specify the `n_iter`
        params = {'n_iter': 50, param_name: param_value}

        partial_embedding = embedding.prepare_partial(self.x_test)
        partial_embedding.optimize(**params, inplace=True)

        self.assertEqual(1, gradient_descent.call_count)
        check_call_contains_kwargs(gradient_descent.mock_calls[0], params)

    @check_params({'metric': set(VALID_METRICS) - {'mahalanobis'}})
    @patch('sklearn.neighbors.NearestNeighbors')
    def test_ball_tree_distances(self, param_name, metric, ball_tree: MagicMock):
        """Distance metrics should be properly passed down to ball tree"""
        assert param_name == 'metric'
        tsne = TSNE(metric=metric, neighbors='exact')

        # We don't care about what happens later, just that the NN method is
        # properly called
        ball_tree.side_effect = InterruptedError()
        try:
            # Haversine distance only supports two dimensions
            tsne.prepare_initial(self.x[:, :2])
        except InterruptedError:
            pass

        self.assertEqual(ball_tree.call_count, 1)
        check_call_contains_kwargs(ball_tree.mock_calls[0], {'metric': metric})

    @patch('sklearn.neighbors.NearestNeighbors')
    def test_ball_tree_mahalanobis_distance(self, ball_tree: MagicMock):
        """Distance metrics and additional params should be correctly passed down to ball tree"""
        metric = 'mahalanobis'
        C = np.cov(self.x)

        tsne = TSNE(metric=metric, metric_params={'V': C}, neighbors='exact')

        # We don't care about what happens later, just that the NN method is
        # properly called
        ball_tree.side_effect = InterruptedError()
        try:
            tsne.prepare_initial(self.x)
        except InterruptedError:
            pass

        self.assertEqual(ball_tree.call_count, 1)
        check_call_contains_kwargs(ball_tree.mock_calls[0], {'metric': metric})

    @check_params({'metric': set(VALID_METRICS) - {'mahalanobis'}})
    @patch('fastTSNE.pynndescent.NNDescent')
    def test_nndescent_distances(self, param_name, metric, nndescent: MagicMock):
        """Distance metrics should be properly passed down to NN descent"""
        assert param_name == 'metric'
        tsne = TSNE(metric=metric, neighbors='approx')

        # We don't care about what happens later, just that the NN method is
        # properly called
        nndescent.side_effect = InterruptedError()
        try:
            # Haversine distance only supports two dimensions
            tsne.prepare_initial(self.x[:, :2])
        except InterruptedError:
            pass

        self.assertEqual(nndescent.call_count, 1)
        check_call_contains_kwargs(nndescent.mock_calls[0], {'metric': metric})

    @patch('fastTSNE.pynndescent.NNDescent')
    def test_nndescent_mahalanobis_distance(self, nndescent: MagicMock):
        """Distance metrics and additional params should be correctly passed down to NN descent"""
        metric = 'mahalanobis'
        C = np.cov(self.x)

        tsne = TSNE(metric=metric, metric_params={'V': C}, neighbors='approx')

        # We don't care about what happens later, just that the NN method is
        # properly called
        nndescent.side_effect = InterruptedError()
        try:
            tsne.prepare_initial(self.x)
        except InterruptedError:
            pass

        self.assertEqual(nndescent.call_count, 1)
        check_call_contains_kwargs(nndescent.mock_calls[0], {'metric': metric})

    def test_raises_error_on_unrecognized_metric(self):
        """Unknown distance metric should raise error"""
        tsne = TSNE(metric='imaginary', neighbors='exact')
        with self.assertRaises(ValueError):
            tsne.prepare_initial(self.x)

        tsne = TSNE(metric='imaginary', neighbors='approx')
        with self.assertRaises(ValueError):
            tsne.prepare_initial(self.x)


class TestTSNEInplaceOptimization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tsne = TSNE()
        cls.x = np.random.randn(100, 4)
        cls.x_test = np.random.randn(25, 4)

    def test_embedding_inplace_optimization(self):
        embedding1 = self.tsne.prepare_initial(self.x)

        embedding2 = embedding1.optimize(n_iter=5, inplace=True)
        embedding3 = embedding2.optimize(n_iter=5, inplace=True)

        self.assertIs(embedding1.base, embedding2.base)
        self.assertIs(embedding2.base, embedding3.base)

    def test_embedding_not_inplace_optimization(self):
        embedding1 = self.tsne.prepare_initial(self.x)

        embedding2 = embedding1.optimize(n_iter=5, inplace=False)
        embedding3 = embedding2.optimize(n_iter=5, inplace=False)

        self.assertFalse(embedding1.base is embedding2.base)
        self.assertFalse(embedding2.base is embedding3.base)
        self.assertFalse(embedding1.base is embedding3.base)

    def test_partial_embedding_inplace_optimization(self):
        # Prepare reference embedding
        embedding = self.tsne.prepare_initial(self.x)
        embedding.optimize(10, inplace=True)

        partial_embedding1 = embedding.prepare_partial(self.x_test)
        partial_embedding2 = partial_embedding1.optimize(5, inplace=True)
        partial_embedding3 = partial_embedding2.optimize(5, inplace=True)

        self.assertIs(partial_embedding1.base, partial_embedding2.base)
        self.assertIs(partial_embedding2.base, partial_embedding3.base)

    def test_partial_embedding_not_inplace_optimization(self):
        # Prepare reference embedding
        embedding = self.tsne.prepare_initial(self.x)
        embedding.optimize(10, inplace=True)

        partial_embedding1 = embedding.prepare_partial(self.x_test)
        partial_embedding2 = partial_embedding1.optimize(5, inplace=False)
        partial_embedding3 = partial_embedding2.optimize(5, inplace=False)

        self.assertFalse(partial_embedding1.base is partial_embedding2.base)
        self.assertFalse(partial_embedding2.base is partial_embedding3.base)
        self.assertFalse(partial_embedding1.base is partial_embedding3.base)


class TestTSNECallbackParams(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tsne = TSNE()
        cls.x = np.random.randn(100, 4)
        cls.x_test = np.random.randn(25, 4)

    def test_can_pass_callbacks_to_tsne_object(self):
        callback = MagicMock()
        callback2 = MagicMock()
        # We don't want individual callbacks to be iterable
        del callback.__iter__
        del callback2.__iter__

        # Should be able to pass a single callback
        TSNE(callbacks=callback, callbacks_every_iters=1,
             early_exaggeration_iter=0, n_iter=1).fit(self.x)
        self.assertEqual(callback.call_count, 1)

        # Should be able to pass a list callbacks
        callback.reset_mock()
        TSNE(callbacks=[callback], callbacks_every_iters=1,
             early_exaggeration_iter=0, n_iter=1).fit(self.x)
        self.assertEqual(callback.call_count, 1)

        # Should be able to change the callback on the object
        callback.reset_mock()
        tsne = TSNE(callbacks=callback, callbacks_every_iters=1,
                    early_exaggeration_iter=0, n_iter=1)
        tsne.callbacks = callback2
        tsne.fit(self.x)
        callback.assert_not_called()
        self.assertEqual(callback2.call_count, 1)

    def test_can_pass_callbacks_to_embedding_optimize(self):
        embedding = self.tsne.prepare_initial(self.x)

        # We don't the callback to be iterable
        callback = MagicMock()
        del callback.__iter__

        # Should be able to pass a single callback
        embedding.optimize(1, callbacks=callback, callbacks_every_iters=1)
        self.assertEqual(callback.call_count, 1)

        # Should be able to pass a list callbacks
        callback.reset_mock()
        embedding.optimize(1, callbacks=[callback], callbacks_every_iters=1)
        self.assertEqual(callback.call_count, 1)

    def test_can_pass_callbacks_to_partial_embedding_optimize(self):
        embedding = self.tsne.prepare_initial(self.x)

        # We don't the callback to be iterable
        callback = MagicMock()
        del callback.__iter__

        # Should be able to pass a single callback
        partial_embedding = embedding.prepare_partial(self.x_test)
        partial_embedding.optimize(1, callbacks=callback, callbacks_every_iters=1)
        self.assertEqual(callback.call_count, 1)

        # Should be able to pass a list callbacks
        callback.reset_mock()
        partial_embedding.optimize(1, callbacks=[callback], callbacks_every_iters=1)
        self.assertEqual(callback.call_count, 1)


class TSNEInitialization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # It would be nice if the initial data were not nicely behaved to test
        # for low variance
        cls.x = np.random.normal(100, 50, (25, 4))

    def test_low_variance(self):
        """Low variance of the initial embedding is very important for the
        convergence of tSNE."""
        # Cycle through various initializations
        initializations = ['random', 'pca']
        allowed = 1e-3

        for init in initializations:
            tsne = TSNE(initialization=init, perplexity=2)
            embedding = tsne.prepare_initial(self.x)
            np.testing.assert_array_less(np.var(embedding, axis=0), allowed,
                                         'using the `%s` initialization' % init)

    def test_mismatching_embedding_dimensions_simple_api(self):
        # Fit
        tsne = TSNE(n_components=2, initialization=self.x[:10, :2])
        with self.assertRaises(ValueError, msg='fit::ncorrect number of points'):
            tsne.fit(self.x[:25])

        with self.assertRaises(ValueError, msg='fit::ncorrect number of dimensions'):
            TSNE(n_components=2, initialization=self.x[:10, :4])

        # Transform
        tsne = TSNE(n_components=2, initialization='random')
        embedding = tsne.fit(self.x)
        with self.assertRaises(ValueError, msg='transform::incorrect number of points'):
            embedding.transform(X=self.x[:5], initialization=self.x[:10, :2])

        with self.assertRaises(ValueError, msg='transform::incorrect number of dimensions'):
            embedding.transform(X=self.x, initialization=self.x[:, :4])


class TestRandomState(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # It would be nice if the initial data were not nicely behaved to test
        # for low variance
        cls.x = np.random.normal(10000, 50, (25, 4))
        cls.x_test = np.random.normal(100, 50, (25, 4))

    def test_same_results_on_fixed_random_state_random_init(self):
        """Results should be exactly the same if we provide a random state."""
        tsne1 = TSNE(random_state=1, initialization='random')
        embedding1 = tsne1.fit(self.x)

        tsne2 = TSNE(random_state=1, initialization='random')
        embedding2 = tsne2.fit(self.x)

        np.testing.assert_array_equal(embedding1, embedding2,
                                      'Same random state produced different initial embeddings')

    def test_same_results_on_fixed_random_state_pca_init(self):
        """Results should be exactly the same if we provide a random state."""
        tsne1 = TSNE(random_state=1, initialization='pca')
        embedding1 = tsne1.fit(self.x)

        tsne2 = TSNE(random_state=1, initialization='pca')
        embedding2 = tsne2.fit(self.x)

        np.testing.assert_array_equal(embedding1, embedding2,
                                      'Same random state produced different initial embeddings')

    def test_same_partial_embedding_on_fixed_random_state(self):
        tsne = TSNE(random_state=1, initialization='random')
        embedding = tsne.fit(self.x)

        partial1 = embedding.prepare_partial(self.x_test, initialization='random')
        partial2 = embedding.prepare_partial(self.x_test, initialization='random')

        np.testing.assert_array_equal(partial1, partial2,
                                      'Same random state produced different partial embeddings')


class TestDefaultParameterSettings(unittest.TestCase):
    def test_default_params_simple_vs_complex_flow(self):
        # Relevant affinity parameters are passed to the affinity object
        mismatching = get_mismatching_default_values(
            TSNE, PerplexityBasedNN, {'neighbors': 'method'})
        self.assertEqual(mismatching, [])

        assert len(get_shared_parameters(tsne.TSNE, tsne.gradient_descent.__call__)) > 0, \
            '`TSNE` and `gradient_descent` have no shared parameters. Have you ' \
            'changed the signature or usage?'

        # The relevant gradient descent parameters are passed down directly to
        # `gradient_descent`
        mismatching = get_mismatching_default_values(tsne.TSNE, tsne.gradient_descent.__call__)
        mismatching = list(filter(lambda x: x[0] not in ('n_iter',), mismatching))
        self.assertEqual(mismatching, [])


def get_shared_parameters(f1, f2):
    """Get the names of shared parameters from two function signatures."""
    params1 = inspect.signature(f1).parameters
    params2 = inspect.signature(f2).parameters

    return set(params1.keys()) & set(params2.keys())


def get_mismatching_default_values(f1, f2, mapping=None):
    """Check that two functions have the same default values for shared parameters."""
    # Additional mappings from f1 parameters to f2 parameters may be provided
    if mapping is None:
        mapping = {}

    params1 = inspect.signature(f1).parameters
    params2 = inspect.signature(f2).parameters

    mismatch = []
    for f1_param_name in params1:
        # If the param is named differently in f2, rename
        f2_param_name = mapping[f1_param_name] if f1_param_name in mapping else f1_param_name

        # If the parameter does not appear in the signature of f2, there's
        # nothing to do
        if f2_param_name not in params2:
            continue

        val1 = params1[f1_param_name].default
        val2 = params2[f2_param_name].default

        if val1 != val2:
            mismatch.append((f1_param_name, val1, f2_param_name, val2))

    return mismatch


class TestGradientDescentOptimizer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tsne = TSNE()
        random_state = np.random.RandomState(42)
        cls.x = random_state.randn(100, 4)
        cls.x_test = random_state.randn(25, 4)

    def test_optimizer_being_passed_to_subsequent_embeddings(self):
        embedding = self.tsne.prepare_initial(self.x)

        self.assertIsNone(embedding.optimizer.gains,
                          'Optimizer should be initialized with no gains')

        # Check the switch from no gains to some gains
        embedding1 = embedding.optimize(10)
        self.assertIsNone(
            embedding.optimizer.gains,
            'Gains changed on initial optimizer even though we did not do '
            'inplace optimization.')
        self.assertIsNotNone(
            embedding1.optimizer.gains,
            'Gains were not properly set in new embedding.')
        self.assertIsNot(
            embedding.optimizer, embedding1.optimizer,
            'The embedding and new embedding optimizer are the same instance '
            'even we did not do inplace optimization.')

        # Check switch from existing gains to new gains
        embedding2 = embedding1.optimize(10)
        self.assertIsNot(
            embedding1.optimizer, embedding2.optimizer,
            'The embedding and new embedding optimizer are the same instance '
            'even we did not do inplace optimization.')
        self.assertFalse(
            np.allclose(embedding1.optimizer.gains, embedding2.optimizer.gains),
            'The gains in the new embedding did not change at all from the old '
            'embedding.'
        )

    def test_optimizer_being_passed_to_partial_embeddings(self):
        embedding = self.tsne.prepare_initial(self.x)
        embedding.optimize(10, inplace=True)

        # Partial embeddings get their own optimizer instance
        partial = embedding.prepare_partial(self.x_test)
        self.assertIsNot(
            embedding.optimizer, partial.optimizer,
            'Embedding and partial embedding optimizers are the same instance.')
        self.assertIsNone(
            partial.optimizer.gains,
            'Partial embedding was not initialized with no gains')

        # Check the switch from no gains to some gains
        partial1 = partial.optimize(10)
        self.assertIsNone(
            partial.optimizer.gains,
            'Gains on initial optimizer changed even though we did not do '
            'inplace optimization.')
        self.assertIsNotNone(
            partial1.optimizer.gains,
            'Gains were not properly set in new partial embedding.')

        # Check switch from existing gains to new gains
        partial2 = partial1.optimize(10)
        self.assertIsNot(
            partial1.optimizer, partial2.optimizer,
            'The embedding and new embedding optimizer are the same instance '
            'even we did not do inplace optimization.')
        self.assertFalse(
            np.allclose(partial1.optimizer.gains, partial2.optimizer.gains),
            'The gains in the new embedding did not change at all from the old '
            'embedding.'
        )

    def test_embedding_optimizer_inplace(self):
        embedding0 = self.tsne.prepare_initial(self.x)

        # Assign only so the references are clear
        embedding1 = embedding0.optimize(10, inplace=True)
        embedding2 = embedding1.optimize(10, inplace=True)

        self.assertIs(embedding0.optimizer, embedding1.optimizer)
        self.assertIs(embedding1.optimizer, embedding2.optimizer)

    def test_partial_embedding_optimizer_inplace(self):
        embedding = self.tsne.prepare_initial(self.x)
        embedding.optimize(10, inplace=True)
        partial0 = embedding.prepare_partial(self.x_test)

        # Assign only so the references are clear
        partial1 = partial0.optimize(10, inplace=True)
        partial2 = partial1.optimize(10, inplace=True)

        self.assertIs(partial0.optimizer, partial1.optimizer)
        self.assertIs(partial1.optimizer, partial2.optimizer)




import logging
import multiprocessing
from collections import Iterable
from types import SimpleNamespace

import numpy as np
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator

from . import _tsne
from . import initialization as initialization_scheme
from .affinity import Affinities, PerplexityBasedNN
from .quad_tree import QuadTree

EPSILON = np.finfo(np.float64).eps

log = logging.getLogger(__name__)


def _check_callbacks(callbacks):
    if callbacks is not None:
        # If list was passed, make sure all of them are actually callable
        if isinstance(callbacks, Iterable):
            if any(not callable(c) for c in callbacks):
                raise ValueError('`callbacks` must contain callable objects!')
        # The gradient descent method deals with lists
        elif callable(callbacks):
            callbacks = (callbacks,)
        else:
            raise ValueError('`callbacks` must be a callable object!')

    return callbacks


def _handle_nice_params(optim_params: dict) -> None:
    """Convert the user friendly params into something the optimizer can
    understand."""
    # Handle callbacks
    optim_params['callbacks'] = _check_callbacks(optim_params.get('callbacks'))
    optim_params['use_callbacks'] = optim_params['callbacks'] is not None

    # Handle negative gradient method
    negative_gradient_method = optim_params.pop('negative_gradient_method')
    if callable(negative_gradient_method):
        negative_gradient_method = negative_gradient_method
    elif negative_gradient_method in {'bh', 'BH', 'barnes-hut'}:
        negative_gradient_method = kl_divergence_bh
    elif negative_gradient_method in {'fft', 'FFT', 'interpolation'}:
        negative_gradient_method = kl_divergence_fft
    else:
        raise ValueError('Unrecognized gradient method. Please choose one of '
                         'the supported methods or provide a valid callback.')
    # `gradient_descent` uses the more informative name `objective_function`
    optim_params['objective_function'] = negative_gradient_method

    # Handle number of jobs
    n_jobs = optim_params['n_jobs']
    if n_jobs < 0:
        n_cores = multiprocessing.cpu_count()
        # Add negative number of n_jobs to the number of cores, but increment by
        # one because -1 indicates using all cores, -2 all except one, and so on
        n_jobs = n_cores + n_jobs + 1

    # If the number of jobs, after this correction is still <= 0, then the user
    # probably thought they had more cores, so we'll default to 1
    if n_jobs <= 0:
        log.warning('`n_jobs` receieved value %d but only %d cores are available. '
                    'Defaulting to single job.' % (optim_params['n_jobs'], n_cores))
        n_jobs = 1

    optim_params['n_jobs'] = n_jobs


def __check_init_num_samples(num_samples, required_num_samples):
    if num_samples != required_num_samples:
        raise ValueError(
            'The provided initialization contains a different number '
            'of points (%d) than the data provided (%d).' % (
                num_samples, required_num_samples)
        )


def __check_init_num_dimensions(num_dimensions, required_num_dimensions):
    if num_dimensions != required_num_dimensions:
        raise ValueError(
            'The provided initialization contains a different number '
            'of components (%d) than the embedding (%d).' % (
                num_dimensions, required_num_dimensions)
        )


init_checks = SimpleNamespace(
    num_samples=__check_init_num_samples,
    num_dimensions=__check_init_num_dimensions,
)


class OptimizationInterrupt(InterruptedError):
    """Optimization was interrupted by a callback

    Parameters
    ----------
    error: float
        The latest KL divergence of the embedding.
    final_embedding: Union[TSNEEmbedding, PartialTSNEEmbedding]
        The latest embedding. Is either a partial or full embedding, depending
        on where the error was raised.

    """
    def __init__(self, error, final_embedding):
        super().__init__()
        self.error = error
        self.final_embedding = final_embedding


class PartialTSNEEmbedding(np.ndarray):
    """A partial embedding is created when we take an existing
    :class:`TSNEEmbedding` and add new data to it. It differs from the typical
    embedding in that it would be unwise to add even more data to only the
    subset of already approximated data.

    If we would like to add new data multiple times to the existing embedding,
    we can simply do so by using the original embedding.

    """

    def __new__(cls, embedding, reference_embedding, P, gradient_descent_params, optimizer=None):
        init_checks.num_samples(embedding.shape[0], P.shape[0])

        obj = np.asarray(embedding, dtype=np.float64, order='C').view(PartialTSNEEmbedding)

        obj.reference_embedding = reference_embedding
        obj.P = P
        obj.gradient_descent_params = gradient_descent_params

        if optimizer is None:
            optimizer = gradient_descent()
        elif not isinstance(optimizer, gradient_descent):
            raise TypeError('`optimizer` must be an instance of `%s`, but got `%s`.' % (
                gradient_descent.__class__.__name__, type(optimizer)))
        obj.optimizer = optimizer

        obj.kl_divergence = None

        return obj

    def optimize(self, n_iter, inplace=False, propagate_exception=False,
                 **gradient_descent_params):
        """Run optmization on the embedding for a given nubmer of steps.

        Parameters
        ----------
        n_iter: int
            The number of optimization iterations.
        inplace: bool
            Whether or not to create a copy of the embedding or to perform
            updates inplace.
        propagate_exception: bool
            The optimization process can be interrupted using callbacks. This
            flag indicates whether we should propagate that exception or to
            simply stop optimization and return the resulting embedding.
        **gradient_descent_params: dict
            Any parameters accepted by :func:`gradient_descent` can be specified
            here for finer control of the optimization process.

        Returns
        -------
        PartialTSNEEmbedding
            An optimized partial t-SNE embedding.

        Raises
        ------
        OptimizationInterrupt
            If a callback stops the optimization and the ``propagate_exception``
            flag is set, then an exception is raised.

        """
        # Typically we want to return a new embedding and keep the old one intact
        if inplace:
            embedding = self
        else:
            embedding = PartialTSNEEmbedding(
                np.copy(self), self.reference_embedding, self.P, self.gradient_descent_params,
                optimizer=self.optimizer.copy(),
            )

        # If optimization parameters were passed to this funciton, prefer those
        # over the defaults specified in the TSNE object
        optim_params = dict(self.gradient_descent_params)
        optim_params.update(gradient_descent_params)
        _handle_nice_params(optim_params)
        optim_params['n_iter'] = n_iter

        try:
            # Run gradient descent with the embedding optimizer so gains are
            # properly updated and kept
            error, embedding = embedding.optimizer(
                embedding=embedding, reference_embedding=self.reference_embedding,
                P=self.P, **optim_params,
            )

        except OptimizationInterrupt as ex:
            log.info('Optimization was interrupted with callback.')
            if propagate_exception:
                raise ex
            error, embedding = ex.error, ex.final_embedding

        embedding.kl_divergence = error

        return embedding


class TSNEEmbedding(np.ndarray):
    """t-SNE embedding

    Parameters
    ----------
    embedding: np.ndarray
        Initial positions for each data point.
    affinities: Affinities
        An affinity index which can be used to compute the affinities of new
        points to the points in the existing embedding. The affinity index also
        contains the affinity matrix :math:`P` used during optimization.
    random_state: Optional[Union[int, RandomState]]
        The random state parameter follows the convention used in scikit-learn.
        If the value is an int, random_state is the seed used by the random
        number generator. If the value is a RandomState instance, then it will
        be used as the random number generator. If the value is None, the random
        number generator is the RandomState instance used by `np.random`.

    Attributes
    ----------
    kl_divergence: float
        The KL divergence or error of the embedding.

    """

    def __new__(cls, embedding, affinities, random_state=None, optimizer=None,
                **gradient_descent_params):
        init_checks.num_samples(embedding.shape[0], affinities.P.shape[0])

        obj = np.asarray(embedding, dtype=np.float64, order='C').view(TSNEEmbedding)

        obj.affinities = affinities  # type: Affinities
        obj.gradient_descent_params = gradient_descent_params  # type: dict
        obj.random_state = random_state

        if optimizer is None:
            optimizer = gradient_descent()
        elif not isinstance(optimizer, gradient_descent):
            raise TypeError('`optimizer` must be an instance of `%s`, but got `%s`.' % (
                gradient_descent.__class__.__name__, type(optimizer)))
        obj.optimizer = optimizer

        obj.kl_divergence = None

        return obj

    def optimize(self, n_iter, inplace=False, propagate_exception=False,
                 **gradient_descent_params):
        """Run optmization on the embedding for a given nubmer of steps.

        Parameters
        ----------
        n_iter: int
            The number of optimization iterations.
        inplace: bool
            Whether or not to create a copy of the embedding or to perform
            updates inplace.
        propagate_exception: bool
            The optimization process can be interrupted using callbacks. This
            flag indicates whether we should propagate that exception or to
            simply stop optimization and return the resulting embedding.
        **gradient_descent_params: dict
            Any parameters accepted by :func:`gradient_descent` can be specified
            here for finer control of the optimization process.

        Returns
        -------
        TSNEEmbedding
            An optimized t-SNE embedding.

        Raises
        ------
        OptimizationInterrupt
            If a callback stops the optimization and the ``propagate_exception``
            flag is set, then an exception is raised.

        """
        # Typically we want to return a new embedding and keep the old one intact
        if inplace:
            embedding = self
        else:
            embedding = TSNEEmbedding(
                np.copy(self), self.affinities, random_state=self.random_state,
                optimizer=self.optimizer.copy(), **self.gradient_descent_params,
            )

        # If optimization parameters were passed to this funciton, prefer those
        # over the defaults specified in the TSNE object
        optim_params = dict(self.gradient_descent_params)
        optim_params.update(gradient_descent_params)
        _handle_nice_params(optim_params)
        optim_params['n_iter'] = n_iter

        try:
            # Run gradient descent with the embedding optimizer so gains are
            # properly updated and kept
            error, embedding = embedding.optimizer(
                embedding=embedding, P=self.affinities.P, **optim_params,
            )

        except OptimizationInterrupt as ex:
            log.info('Optimization was interrupted with callback.')
            if propagate_exception:
                raise ex
            error, embedding = ex.error, ex.final_embedding

        embedding.kl_divergence = error

        return embedding

    def transform(self, X, perplexity=None, initialization='median',
                  early_exaggeration=2, early_exaggeration_iter=100,
                  initial_momentum=0.2, n_iter=100, final_momentum=0.4):
        """Embed new points into the existing embedding.

        This procedure optimizes each point only with respect to the existing
        embedding i.e. it ignores any interactions between the points in ``X``
        among themselves.

        Parameters
        ----------
        X: np.ndarray
            The data matrix to be added to the existing embedding.
        perplexity: float
            Perplexity can be thought of as the continuous k number of neighbors
            to consider for each data point. To avoid confusion, note that
            perplexity linearly impacts runtime.
        initialization: Union[np.ndarray, str]
            The initial point positions to be used in the embedding space. Can
            be a precomputed numpy array, ``random`` or ``weighted``. In all
            cases, ``weighted`` should be preferred. It positions each point in
            the weighted mean position of it's nearest neighbors in the existing
            embedding. Typically, few optimization steps are needed for good
            embeddings. ``random`` positions all new points in the center of the
            embedding and should be used for demonstration only.
        early_exaggeration: float
            The early exaggeration factor.
        early_exaggeration_iter: int
            The number of iterations to run in the early exaggeration phase.
        initial_momentum: float
            As in regular t-SNE, optimization uses momentum for faster
            convergence. This value controls the momentum used during the
            *early exaggeration* phase.
        n_iter: int
            The number of iterations to run in the normal optimization regime.
        final_momentum: float
            As in regular t-SNE, optimization uses momentum for faster
            convergence. This value controls the momentum used during the normal
            regime and*late exaggeration* phase.

        Returns
        -------
        PartialTSNEEmbedding
            The positions of the new points in the embedding space.

        """
        embedding = self.prepare_partial(X, perplexity=perplexity, initialization=initialization)

        try:
            # Early exaggeration with lower momentum to allow points to find more
            # easily move around and find their neighbors
            embedding.optimize(
                n_iter=early_exaggeration_iter, exaggeration=early_exaggeration,
                momentum=initial_momentum, inplace=True, propagate_exception=True,
            )

            # Restore actual affinity probabilities and increase momentum to get
            # final, optimized embedding
            embedding.optimize(
                n_iter=n_iter, exaggeration=None, momentum=final_momentum,
                inplace=True, propagate_exception=True,
            )

        except OptimizationInterrupt as ex:
            log.info('Optimization was interrupted with callback.')
            embedding = ex.final_embedding

        return embedding

    def prepare_partial(self, X, initialization='median', **affinity_params):
        """Prepare the partial embedding which can be optimized.

        Parameters
        ----------
        X : np.ndarray
            The data matrix to be added to the existing embedding.
        initialization: Union[np.ndarray, str]
            The initial point positions to be used in the embedding space. Can
            be a precomputed numpy array, ``random``, ``weighted`` or ``median``.
            In all cases, ``weighted`` or ``median`` should be preferred. These
            positions each point at the weighted mean or median position of it's
            nearest neighbors in the existing embedding. Typically, few
            optimization steps are needed for good embeddings. ``weighted`` and
            ``median`` produce very similar results.

            ``random`` positions all new points in the center of the embedding
            and should be used for demonstration only.
        **affinity_params: dict
            Additional params to be passed to the ``Affinity.to_new`` method.

        Returns
        -------
        PartialTSNEEmbedding
            An unoptimized :class:`PartialTSNEEmbedding` object, prepared for
            optimization.

        """
        P = self.affinities.to_new(X, **affinity_params)

        # Extract neighbor indices and their affinities as a dense matrix from P
        # so they can be used in weighted initialization schemes
        P = P.tocsr()
        k_neighbors = int(P.data.shape[0] / P.shape[0])
        neighbors = P.indices.copy().reshape((P.shape[0], k_neighbors))
        distances = P.data.copy().reshape((P.shape[0], k_neighbors))

        # If initial positions are given in an array, use a copy of that
        if isinstance(initialization, np.ndarray):
            init_checks.num_samples(initialization.shape[0], X.shape[0])
            init_checks.num_dimensions(initialization.shape[1], self.shape[1])

            embedding = np.array(initialization)

        # Random initialization with isotropic normal distribution
        elif initialization == 'random':
            embedding = initialization_scheme.random(X.shape[0], self.shape[1], self.random_state)
        elif initialization == 'weighted':
            embedding = initialization_scheme.weighted_mean(X, self, neighbors, distances)
        elif initialization == 'median':
            embedding = initialization_scheme.median(self, neighbors)
        else:
            raise ValueError('Unrecognized initialization scheme `%s`.' % initialization)

        return PartialTSNEEmbedding(
            embedding, reference_embedding=self, P=P,
            gradient_descent_params=self.gradient_descent_params,
        )


class TSNE(BaseEstimator):
    """t-Distributed Stochastic Neighbor Embedding

    Parameters
    ----------
    n_components: int
        The dimension of the embedding space.
    perplexity: float
        Perplexity can be thought of as the continuous :math:`k` number of
        neighbors to consider for each data point. To avoid confusion, note that
        perplexity linearly impacts runtime.
    learning_rate: float
        The learning rate for the t-SNE optimization steps. Typical values range
        from 1 to 1000. Setting the learning rate too low or too high may result
        in the points forming a "ball". This is also known as the crowding
        problem.
    early_exaggeration_iter: int
        The number of iterations to run in the *early exaggeration* phase.
    early_exaggeration: float
        The early exaggeration factor. Typical values range from 12 to 32,
        however larger values have also been found to work well with specific
        values of learning rate. See Linderman and Steinberger [3]_ for more
        details.
    n_iter: int
        The number of iterations to run in the normal optimization regime.
    exaggeration: Optional[int]
        The exaggeration factor to be used during the normal optmimization
        phase. Standard implementation don't use this exaggeration and it
        typically isn't necessary for smaller data sets, but it has been shown
        that for larger data sets, using some exaggeration is necessary in order
        to obtain good embeddings.
    theta: float
        Only used when ``negative_gradient_method='bh'`` or its other aliases.
        This is the trade-off parameter between speed and accuracy of the tree
        approximation method. Typical values range from 0.2 to 0.8. The value 0
        indicates that no approximation is to be made and produces exact results
        also producing longer runtime. See [2]_ for more details.
    n_interpolation_points: int
        Only used when ``negative_gradient_method='fft'`` or its other aliases.
        The number of interpolation points to use within each grid cell for
        interpolation based t-SNE. It is highly recommended leaving this value
        at the default 3 as otherwise the interpolation may suffer from the
        Runge phenomenon. Theoretically, increasing this number will result in
        higher approximation accuracy, but practically, this can also be done
        with the ``ints_in_interval`` parameter, which does not suffer from the
        Runge phenomenon and should always be preferred. This is described in
        detail by Linderman [2]_.
    min_num_intervals: int
        Only used when ``negative_gradient_method='fft'`` or its other aliases.
        The interpolation approximation method splits the embedding space into a
        grid, where the number of grid cells is governed by
        ``ints_in_interval``. Sometimes, especially during early stages of
        optimization, that number may be too small, and we may want better
        accuracy. The number of intervals used will thus always be at least the
        number specified here. Note that larger values will produce more precise
        approximations but will have longer runtime.
    ints_in_interval: float
        Only used when ``negative_gradient_method='fft'`` or its other aliases.
        Since the coordinate range of the embedding changes during optimization,
        this value tells us how many integers should appear in a single e.g.
        setting this value to 3 means that the intervals will appear as follows:
        [0, 3), [3, 6), ... Lower values will need more intervals to fill the
        space, e.g. 1.5 will produce 4 intervals [0, 1.5), [1.5, 3), ...
        Therefore lower values will produce more intervals, producing more
        interpolation points which in turn produce better approximation at the
        cost of longer runtime.
    initialization: Union[np.ndarray, str]
        The initial point positions to be used in the embedding space. Can be a
        precomputed numpy array, ``pca`` or ``random``. Please note that when
        passing in a precomputed positions, it is highly recommended that the
        point positions have small variance (var(Y) < 0.0001), otherwise you may
        get poor embeddings.
    metric: str
        The metric to be used to compute affinities between points in the
        original space.
    metric_params: Optional[dict]
        Additional keyword arguments for the metric function.
    initial_momentum: float
        t-SNE optimization uses momentum for faster convergence. This value
        controls the momentum used during the *early optimization* phase.
    final_momentum: float
        t-SNE optimization uses momentum for faster convergence. This value
        controls the momentum used during the normal regime and *late
        exaggeration* phase.
    n_jobs: int
        The number of jobs to run in parallel. This follows the scikit-learn
        convention, ``-1`` meaning all processors, ``-2`` meaning all but one
        processor and so on.
    neighbors: str
        Specifies the nearest neighbor method to use. Can be either ``exact`` or
        ``approx``. ``exact`` uses space partitioning binary trees from
        scikit-learn while ``approx`` makes use of nearest neighbor descent.
        Note that ``approx`` has a bit of overhead and will be slower on smaller
        data sets than exact search.
    negative_gradient_method: str
        Specifies the negative gradient approximation method to use. For smaller
        data sets, the Barnes-Hut approximation is appropriate [2]_ and can be
        set using one of the following aliases: ``bh``, ``BH`` or
        ``barnes-hut``. Note that the time complexity of Barnes-Hut scales as
        :math:`\mathcal{O}(N \log N)`. For larger data sets, the FFT accelerated
        interpolation method is more appropriate and can be set using one of the
        following aliases: ``fft``, ``FFT`` or ``ìnterpolation`` [4]_. Note that
        this method scales linearly in the number of points
        :math:`\mathcal{O}(N)` and its complexity is governed by the number of
        interpolation points.
    callbacks: Optional[Union[Callable, List[Callable]]]
        We can pass callbacks, that will be run every ``callbacks_every_iters``
        iterations. Each callback should accept three parameters, the first is
        the current iteration number, the second is the current KL divergence
        error and the last is the current embedding. The callback may return
        ``True`` in order to stop the optimization.
    callbacks_every_iters: int
        How many iterations should run between each time a callback is invoked.
    random_state: Optional[Union[int, RandomState]]
        The random state parameter follows the convention used in scikit-learn.
        If the value is an int, random_state is the seed used by the random
        number generator. If the value is a RandomState instance, then it will
        be used as the random number generator. If the value is None, the random
        number generator is the RandomState instance used by `np.random`.

    References
    ----------
    .. [1] Maaten, Laurens van der, and Geoffrey Hinton. "Visualizing data using
        t-SNE." Journal of machine learning research 9.Nov (2008): 2579-2605.
    .. [2] Van Der Maaten, Laurens. "Accelerating t-SNE using tree-based
        algorithms." The Journal of Machine Learning Research 15.1 (2014):
        3221-3245.
    .. [3] Linderman, George C., and Stefan Steinerberger. "Clustering with
        t-SNE, provably." arXiv preprint arXiv:1706.02582 (2017).
    .. [4] Linderman, George C., et al. "Efficient Algorithms for t-distributed
        Stochastic Neighborhood Embedding." arXiv preprint arXiv:1712.09005
        (2017).

    """

    def __init__(self, n_components=2, perplexity=30, learning_rate=100,
                 early_exaggeration_iter=250, early_exaggeration=12,
                 n_iter=750, exaggeration=None,
                 theta=0.5, n_interpolation_points=3, min_num_intervals=10,
                 ints_in_interval=1, initialization='pca', metric='euclidean',
                 metric_params=None, initial_momentum=0.5, final_momentum=0.8,
                 n_jobs=1, neighbors='exact', negative_gradient_method='bh',
                 callbacks=None, callbacks_every_iters=50, random_state=None):
        self.n_components = n_components
        self.perplexity = perplexity
        self.learning_rate = learning_rate
        self.early_exaggeration = early_exaggeration
        self.early_exaggeration_iter = early_exaggeration_iter
        self.n_iter = n_iter
        self.exaggeration = exaggeration
        self.theta = theta
        self.n_interpolation_points = n_interpolation_points
        self.min_num_intervals = min_num_intervals
        self.ints_in_interval = ints_in_interval

        # Check if the number of components match the initialization dimension
        if isinstance(initialization, np.ndarray):
            init_checks.num_dimensions(initialization.shape[1], n_components)
        self.initialization = initialization

        self.metric = metric
        self.metric_params = metric_params
        self.initial_momentum = initial_momentum
        self.final_momentum = final_momentum
        self.n_jobs = n_jobs
        self.neighbors_method = neighbors
        self.negative_gradient_method = negative_gradient_method

        self.callbacks = callbacks
        self.callbacks_every_iters = callbacks_every_iters

        self.random_state = random_state

    def fit(self, X):
        """Fit a t-SNE embedding for a given data set.

        Runs the standard t-SNE optimization, consisting of the early
        exaggeration phase and a normal optimization phase.

        Parameters
        ----------
        X: np.ndarray
            The data matrix to be embedded.

        Returns
        -------
        TSNEEmbedding
            A fully optimized t-SNE embedding.

        """
        embedding = self.prepare_initial(X)

        try:
            # Early exaggeration with lower momentum to allow points to find more
            # easily move around and find their neighbors
            embedding.optimize(
                n_iter=self.early_exaggeration_iter, exaggeration=self.early_exaggeration,
                momentum=self.initial_momentum, inplace=True, propagate_exception=True,
            )

            # Restore actual affinity probabilities and increase momentum to get
            # final, optimized embedding
            embedding.optimize(
                n_iter=self.n_iter, exaggeration=self.exaggeration,
                momentum=self.final_momentum, inplace=True, propagate_exception=True,
            )

        except OptimizationInterrupt as ex:
            log.info('Optimization was interrupted with callback.')
            embedding = ex.final_embedding

        return embedding

    def prepare_initial(self, X):
        """Prepare the initial embedding which can be optimized as needed.

        Parameters
        ----------
        X: np.ndarray
            The data matrix to be embedded.

        Returns
        -------
        TSNEEmbedding
            An unoptimized :class:`TSNEEmbedding` object, prepared for
            optimization.

        """
        # If initial positions are given in an array, use a copy of that
        if isinstance(self.initialization, np.ndarray):
            init_checks.num_samples(self.initialization.shape[0], X.shape[0])
            init_checks.num_dimensions(self.initialization.shape[1], self.n_components)

            embedding = np.array(self.initialization)

            variance = np.var(embedding, axis=0)
            if any(variance > 1e-4):
                log.warning(
                    'Variance of embedding is greater than 0.0001. Initial '
                    'embeddings with high variance may have display poor convergence.'
                )

        elif self.initialization == 'pca':
            embedding = initialization_scheme.pca(
                X, self.n_components, scale_down=True, random_state=self.random_state,
            )
        elif self.initialization == 'random':
            embedding = initialization_scheme.random(
                X.shape[0], self.n_components, random_state=self.random_state,
            )
        else:
            raise ValueError('Unrecognized initialization scheme `%s`.' % self.initialization)

        affinities = PerplexityBasedNN(
            X, self.perplexity, method=self.neighbors_method,
            metric=self.metric, metric_params=self.metric_params, n_jobs=self.n_jobs,
        )

        gradient_descent_params = {
            # Degrees of freedom of the Student's t-distribution. The
            # suggestion degrees_of_freedom = n_components - 1 comes from [3]_.
            'dof':  max(self.n_components - 1, 1),

            'negative_gradient_method': self.negative_gradient_method,
            'learning_rate': self.learning_rate,
            # By default, use the momentum used in unexaggerated phase
            'momentum': self.final_momentum,

            # Barnes-Hut params
            'theta': self.theta,
            # Interpolation params
            'n_interpolation_points': self.n_interpolation_points,
            'min_num_intervals': self.min_num_intervals,
            'ints_in_interval': self.ints_in_interval,

            'n_jobs': self.n_jobs,
            # Callback params
            'callbacks': self.callbacks,
            'callbacks_every_iters': self.callbacks_every_iters,
        }

        return TSNEEmbedding(embedding, affinities, self.random_state, **gradient_descent_params)


def kl_divergence_bh(embedding, P, dof, bh_params, reference_embedding=None,
                     should_eval_error=False, n_jobs=1, **_):
    gradient = np.zeros_like(embedding, dtype=np.float64, order='C')

    # In the event that we wish to embed new points into an existing embedding
    # using simple optimization, we compute optimize the new embedding points
    # w.r.t. the existing embedding. Otherwise, we want to optimize the
    # embedding w.r.t. itself
    if reference_embedding is None:
        reference_embedding = embedding

    # Compute negative gradient
    tree = QuadTree(reference_embedding)
    sum_Q = _tsne.estimate_negative_gradient_bh(
        tree, embedding, gradient, **bh_params, dof=dof, num_threads=n_jobs)
    del tree

    # Compute positive gradient
    sum_P, kl_divergence_ = _tsne.estimate_positive_gradient_nn(
        P.indices, P.indptr, P.data, embedding, reference_embedding, gradient,
        dof, num_threads=n_jobs, should_eval_error=should_eval_error,
    )

    gradient *= 2 * (dof + 1) / dof
    # Computing positive gradients summed up only unnormalized q_ijs, so we
    # have to include normalziation term separately
    if should_eval_error:
        kl_divergence_ += sum_P * np.log(sum_Q + EPSILON)

    return kl_divergence_, gradient


def kl_divergence_fft(embedding, P, dof, fft_params, reference_embedding=None,
                      should_eval_error=False, n_jobs=1, **_):
    gradient = np.zeros_like(embedding, dtype=np.float64, order='C')

    # Compute negative gradient.
    if embedding.ndim == 1 or embedding.shape[1] == 1:
        if reference_embedding is not None:
            sum_Q = _tsne.estimate_negative_gradient_fft_1d_with_reference(
                embedding.ravel(), reference_embedding.ravel(), gradient.ravel(), **fft_params)
        else:
            sum_Q = _tsne.estimate_negative_gradient_fft_1d(
                embedding.ravel(), gradient.ravel(), **fft_params)
    elif embedding.shape[1] == 2:
        if reference_embedding is not None:
            sum_Q = _tsne.estimate_negative_gradient_fft_2d_with_reference(
                embedding, reference_embedding, gradient, **fft_params)
        else:
            sum_Q = _tsne.estimate_negative_gradient_fft_2d(
                embedding, gradient, **fft_params)
    else:
        raise RuntimeError('Interpolation based t-SNE for >2 dimensions is '
                           'currently unsupported (and generally a bad idea)')

    # The positive gradient function needs a reference embedding always
    if reference_embedding is None:
        reference_embedding = embedding

    # Compute positive gradient
    sum_P, kl_divergence_ = _tsne.estimate_positive_gradient_nn(
        P.indices, P.indptr, P.data, embedding, reference_embedding, gradient,
        dof, num_threads=n_jobs, should_eval_error=should_eval_error,
    )

    gradient *= 2 * (dof + 1) / dof
    if should_eval_error:
        kl_divergence_ += sum_P * np.log(sum_Q + EPSILON)

    return kl_divergence_, gradient


class gradient_descent:
    def __init__(self):
        self.gains = None

    def copy(self):
        optimizer = self.__class__()
        if self.gains is not None:
            optimizer.gains = np.copy(self.gains)
        return optimizer

    def __call__(self, embedding, P, n_iter, objective_function, learning_rate=100,
                 momentum=0.5, exaggeration=None, dof=1, min_gain=0.01,
                 min_grad_norm=1e-8, theta=0.5, n_interpolation_points=3,
                 min_num_intervals=10, ints_in_interval=1, reference_embedding=None,
                 n_jobs=1, use_callbacks=False, callbacks=None, callbacks_every_iters=50):
        """Perform batch gradient descent with momentum and gains.

        Parameters
        ----------
        embedding: np.ndarray
            The embedding :math:`Y`.
        P: csr_matrix
            Joint probability matrix :math:`P`.
        n_iter: int
            The number of iterations to run for.
        objective_function: Callable[..., Tuple[float, np.ndarray]]
            A callable that evaluates the error and gradient for the current
            embedding.
        learning_rate: float
            The learning rate for the t-SNE optimization steps. Typical values range
            from 1 to 1000. Setting the learning rate too low or too high may result
            in the points forming a "ball". This is also known as the crowding
            problem.
        momentum: float
            Momentum accounts for gradient directions from previous iterations and
            resulting in faster convergence.
        exaggeration: float
            The exaggeration term is used to increase the attractive forces of
            nearby points.
        dof: float
            Degrees of freedom of the Student's t-distribution.
        min_gain: float
            Minimum individual gain for each parameter.
        min_grad_norm: float
            If the gradient norm is below this threshold, the optimization will be
            stopped.
        theta: float
            This is the trade-off parameter between speed and accuracy of the
            Barnes-Hut approximation of the negative forces. Setting a lower value
            will produce more accurate results, while setting a higher value will
            search through less of the space providing a rougher approximation.
            Scikit-learn recommends values between 0.2-0.8. This value is ignored
            unless the Barnes-Hut algorithm is used for gradients.
        n_interpolation_points: int
            The number of interpolation points to use for FFT accelerated
            interpolation based tSNE. It is recommended leaving this value at the
            default=3 as otherwise the interpolation may suffer from the Runge
            phenomenon. This value is ignored unless the interpolation based
            algorithm is used.
        min_num_intervals: int
            The minimum number of intervals into which we split our embedding. A
            larger value will produce better embeddings at the cost of performance.
            This value is ignored unless the interpolation based algorithm is used.
        ints_in_interval: float
            Since the coordinate range of the embedding will certainly change
            during optimization, this value tells us how many integer values should
            appear in a single interval. This number of intervals affect the
            embedding quality at the cost of performance. Less ints per interval
            will incur a larger number of intervals.
        reference_embedding: Optional[np.ndarray]
            If we are adding points to an existing embedding, we have to compute
            the gradients and errors w.r.t. the existing embedding.
        n_jobs: int
            Number of threads.
        use_callbacks: bool
        callbacks: Callable[[int, float, np.ndarray] -> bool]
            The callback should accept three parameters, the first is the current
            iteration, the second is the current KL divergence error and the last
            is the current embedding. The callback should return a boolean value
            indicating whether or not to stop optimization i.e. True to stop.
        callbacks_every_iters: int
            How often should the callback be called.

        Returns
        -------
        float
            The KL divergence of the optimized embedding.
        np.ndarray
            The optimized embedding Y.

        Raises
        ------
        OptimizationInterrupt
            If the provided callback interrupts the optimization, this is raised.

        """
        assert isinstance(embedding, np.ndarray), \
            '`embedding` must be an instance of `np.ndarray`. Got `%s` instead' \
            % type(embedding)

        if reference_embedding is not None:
            assert isinstance(reference_embedding, np.ndarray), \
                '`reference_embedding` must be an instance of `np.ndarray`. Got ' \
                '`%s` instead' % type(reference_embedding)

        update = np.zeros_like(embedding)
        if self.gains is None:
            self.gains = np.ones_like(embedding)

        bh_params = {'theta': theta}
        fft_params = {'n_interpolation_points': n_interpolation_points,
                      'min_num_intervals': min_num_intervals,
                      'ints_in_interval': ints_in_interval}

        # Lie about the P values for bigger attraction forces
        if exaggeration is None:
            exaggeration = 1

        if exaggeration != 1:
            P *= exaggeration

        # Notify the callbacks that the optimization is about to start
        if isinstance(callbacks, Iterable):
            for callback in callbacks:
                # Only call function if present on object
                getattr(callback, 'optimzation_about_to_start', lambda: ...)()

        for iteration in range(n_iter):
            should_call_callback = use_callbacks and (iteration + 1) % callbacks_every_iters == 0
            should_eval_error = should_call_callback

            error, gradient = objective_function(
                embedding, P, dof=dof, bh_params=bh_params, fft_params=fft_params,
                reference_embedding=reference_embedding, n_jobs=n_jobs,
                should_eval_error=should_eval_error,
            )

            # Correct the KL divergence w.r.t. the exaggeration if needed
            if should_eval_error and exaggeration != 1:
                error = error / exaggeration - np.log(exaggeration)

            if should_call_callback:
                # Continue only if all the callbacks say so
                should_stop = any((bool(c(iteration + 1, error, embedding)) for c in callbacks))
                if should_stop:
                    # Make sure to un-exaggerate P so it's not corrupted in future runs
                    if exaggeration != 1:
                        P /= exaggeration
                    raise OptimizationInterrupt(error=error, final_embedding=embedding)

            # Update the embedding using the gradient
            grad_direction_flipped = np.sign(update) != np.sign(gradient)
            grad_direction_same = np.invert(grad_direction_flipped)
            self.gains[grad_direction_flipped] += 0.2
            self.gains[grad_direction_same] = self.gains[grad_direction_same] * 0.8 + min_gain
            update = momentum * update - learning_rate * self.gains * gradient
            embedding += update

            # Zero-mean the embedding
            embedding -= np.mean(embedding, axis=0)

            if np.linalg.norm(gradient) < min_grad_norm:
                log.info('Gradient norm eps reached. Finished.')
                break

        # Make sure to un-exaggerate P so it's not corrupted in future runs
        if exaggeration != 1:
            P /= exaggeration

        # The error from the loop is the one for the previous, non-updated
        # embedding. We need to return the error for the actual final embedding, so
        # compute that at the end before returning
        error, _ = objective_function(
            embedding, P, dof=dof, bh_params=bh_params, fft_params=fft_params,
            reference_embedding=reference_embedding, n_jobs=n_jobs,
            should_eval_error=True,
        )

        return error, embedding



import numpy as np
from sklearn.decomposition import PCA
from sklearn.utils import check_random_state


def random(n_samples, n_components, random_state=None):
    """Initialize an embedding using samples from an isotropic Gaussian.

    Parameters
    ----------
    n_samples: int
    n_components: int
    random_state:

    Returns
    -------
    np.ndarray

    """
    random_state = check_random_state(random_state)
    return random_state.normal(0, 1e-2, (n_samples, n_components))


def pca(X, n_components=2, scale_down=True, random_state=None):
    """Initialize am embedding using the top principal components.

    Parameters
    ----------
    X: np.ndarray
    n_components: int
    scale_down: bool
    random_state:

    Returns
    -------
    np.ndarray

    """
    pca_ = PCA(n_components=n_components, random_state=random_state)
    embedding = pca_.fit_transform(X)

    # The PCA embedding may have high variance, which leads to poor convergence
    if scale_down:
        normalization = np.std(embedding[:, 0]) * 100
        embedding /= normalization

    return embedding


def weighted_mean(X, embedding, neighbors, distances):
    """Initialize points onto an existing embedding by placing them in the
    weighted mean position of their nearest neighbors on the reference embedding.

    Parameters
    ----------
    X: np.ndarray
    embedding: TSNEEmbedding
    neighbors: np.ndarray
    distances: np.ndarray

    Returns
    -------
    np.ndarray

    """
    n_samples = X.shape[0]
    n_components = embedding.shape[1]

    partial_embedding = np.zeros((n_samples, n_components))
    for i in range(n_samples):
        partial_embedding[i] = np.average(
            embedding[neighbors[i]], axis=0, weights=distances[i],
        )

    return partial_embedding


def median(embedding, neighbors):
    """Initialize points onto an existing embedding by placing them in the
    median position of their nearest neighbors on the reference embedding.

    Parameters
    ----------
    embedding: TSNEEmbedding
    neighbors: np.ndarray

    Returns
    -------
    np.ndarray

    """
    return np.median(embedding[neighbors], axis=1)



import sys

import numpy as np
from sklearn import neighbors
from sklearn.utils import check_random_state

# In case we're running on a 32bit system, we have to properly handle numba's
# ``parallel`` directive, which throws a ``RuntimeError``. It is important to
# patch this before importing ``pynndescent`` which heavily relies on numba
uns1 = sys.platform.startswith('win32') and sys.version_info[:2] == (2, 7)
uns2 = sys.maxsize <= 2 ** 32
if uns1 or uns2:
    import numba

    __njit_copy = numba.njit

    # Ignore njit decorator and run raw Python function
    def __njit_wrapper(*args, **kwargs):
        return lambda f: f

    numba.njit = __njit_wrapper

    from . import pynndescent

    pynndescent.pynndescent_.numba.njit = __njit_wrapper
    pynndescent.distances.numba.njit = __njit_wrapper
    pynndescent.rp_trees.numba.njit = __njit_wrapper
    pynndescent.utils.numba.njit = __njit_wrapper

from . import pynndescent

# To keep things simple and consistent, we'll only support distances that are
# included in both exact and approximation nearest neighbor search libraries
__ball_tree_metrics = set(neighbors.BallTree.valid_metrics)
__nndescent_metrics = set(pynndescent.distances.named_distances)
VALID_METRICS = sorted(list(__ball_tree_metrics & __nndescent_metrics))


class KNNIndex:
    def __init__(self, metric, metric_params=None, n_jobs=1, random_state=None):
        self.index = None
        self.metric = metric
        self.metric_params = metric_params
        self.n_jobs = n_jobs
        self.random_state = random_state

    def build(self, data):
        """Build the index so we can query nearest neighbors."""

    def query_train(self, data, k):
        """Query the index for the points used to build index."""

    def query(self, query, k):
        """Query the index with new points."""


class BallTree(KNNIndex):
    def build(self, data):
        self.index = neighbors.NearestNeighbors(
            algorithm='ball_tree', metric=self.metric,
            metric_params=self.metric_params, n_jobs=self.n_jobs,
        )
        self.index.fit(data)

    def query_train(self, data, k):
        distances, neighbors = self.index.kneighbors(n_neighbors=k)
        return neighbors, distances

    def query(self, query, k):
        distances, neighbors = self.index.kneighbors(query, n_neighbors=k)
        return neighbors, distances


class NNDescent(KNNIndex):
    def build(self, data):
        random_state = check_random_state(self.random_state)

        # These values were taken from UMAP, which we assume to be sensible defaults
        n_trees = 5 + int(round((data.shape[0]) ** 0.5 / 20))
        n_iters = max(5, int(round(np.log2(data.shape[0]))))

        self.index = pynndescent.NNDescent(
            data, metric=self.metric, metric_kwds=self.metric_params,
            random_state=random_state, n_trees=n_trees, n_iters=n_iters,
            algorithm='alternative', max_candidates=60,
        )

    def query_train(self, data, k):
        search_neighbors = min(data.shape[0] - 1, k + 1)
        neighbors, distances = self.index.query(data, k=search_neighbors, queue_size=1)
        return neighbors[:, 1:], distances[:, 1:]

    def query(self, query, k):
        return self.index.query(query, k=k, queue_size=1)



import unittest

import numpy as np

from fastTSNE.affinity import Multiscale, PerplexityBasedNN


class TestPerplexityBased(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.x = np.random.normal(100, 50, (91, 4))

    def test_properly_reduces_large_perplexity(self):
        aff = PerplexityBasedNN(self.x, perplexity=40)
        self.assertEqual(aff.perplexity, 30)

    def test_handles_reducing_perplexity_value(self):
        perplexity = 20
        k_neighbors = perplexity * 3
        aff = PerplexityBasedNN(self.x, perplexity=perplexity)

        self.assertEqual(aff.perplexity, perplexity)

        # Check that the initial `P` matrix is allright
        n_samples = self.x.shape[0]
        original_P = aff.P.copy()
        # Can't check for equality because the matrix is symmetrized therefore
        # each point may have non-zero values in more than just the k neighbors
        self.assertTrue(original_P.nnz >= n_samples * k_neighbors)

        # Check that lowering the perplexity properly changes affinity matrix
        perplexity = 10
        k_neighbors = perplexity * 3

        aff.set_perplexity(perplexity)
        self.assertEqual(aff.perplexity, perplexity)
        reduced_P = aff.P.copy()
        self.assertTrue(reduced_P.nnz >= n_samples * k_neighbors)
        self.assertTrue(reduced_P.nnz < original_P.nnz,
                        'Lower perplexities should consider less neighbors, '
                        'resulting in a sparser affinity matrix')

        # Raising the perplexity above the initial value would need to recompute
        # the nearest neighbors, so it should raise an error
        with self.assertRaises(RuntimeError):
            aff.set_perplexity(30)


class TestMultiscale(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.x = np.random.normal(100, 50, (91, 4))

    def test_handles_too_large_perplexities(self):
        # x has 91 samples, this means that the max perplexity that we allow is
        # (91 - 1) / 3 = 30. -1 because we don't consider ith point. Anything
        # above that should be ignored or corrected

        ms = Multiscale(self.x, perplexities=[20])
        np.testing.assert_array_equal(
            ms.perplexities, [20],
            'Incorrectly changed perplexity that was within a valid range',
        )

        ms = Multiscale(self.x, perplexities=[20, 40])
        np.testing.assert_array_equal(
            ms.perplexities, [20, 30],
            'Did not lower large perplexity.'
        )

        ms = Multiscale(self.x, perplexities=[20, 40, 60])
        np.testing.assert_array_equal(
            ms.perplexities, [20, 30],
            'Did not drop large perplexities when more than one was too large.'
        )

        ms = Multiscale(self.x, perplexities=[20, 30, 40, 60])
        np.testing.assert_array_equal(
            ms.perplexities, [20, 30],
            'Did not drop duplicate corrected perplexity.'
        )



import gzip
import os
import pickle
import time
import urllib
from os.path import abspath, dirname, join

import fire
import matplotlib.pyplot as plt
import numpy as np
from MulticoreTSNE import MulticoreTSNE
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE as SKLTSNE
from sklearn.model_selection import train_test_split

from fastTSNE.callbacks import ErrorLogger, ErrorApproximations
from fastTSNE.tsne import TSNE, TSNEEmbedding


FILE_DIR = dirname(abspath(__file__))
DATA_DIR = join(FILE_DIR, 'data')

np.set_printoptions(precision=4, suppress=True)
np.random.seed(42)


def plot(x: np.ndarray, y: np.ndarray, show: bool = True, **kwargs) -> None:
    for yi in np.unique(y):
        mask = y == yi
        plt.plot(x[mask, 0], x[mask, 1], 'o', label=str(yi),
                 alpha=kwargs.get('alpha', 0.6), ms=kwargs.get('ms', 1))
    plt.legend()
    if show:
        plt.show()


def plot1d(x: np.ndarray, y: np.ndarray, show: bool = True, **kwargs) -> None:
    for yi in np.unique(y):
        mask = y == yi
        jitter = np.random.randn(mask.shape[0])
        plt.plot(x, jitter, 'o', label=str(yi),
                 alpha=kwargs.get('alpha', 0.6), ms=kwargs.get('ms', 1))
    plt.legend()
    if show:
        plt.show()


def get_mnist(n_samples=None):
    if not os.path.exists(join(DATA_DIR, 'mnist.pkl.gz')):
        urllib.request.urlretrieve(
            'http://deeplearning.net/data/mnist/mnist.pkl.gz', join(DATA_DIR, 'mnist.pkl.gz'))

    with gzip.open(join(DATA_DIR, 'mnist.pkl.gz'), 'rb') as f:
        train, val, test = pickle.load(f, encoding='latin1')
    _train = np.asarray(train[0], dtype=np.float64)
    _val = np.asarray(val[0], dtype=np.float64)
    _test = np.asarray(test[0], dtype=np.float64)
    x = np.vstack((_train, _val, _test))
    y = np.hstack((train[1], val[1], test[1]))

    if n_samples is not None:
        indices = np.random.choice(list(range(x.shape[0])), n_samples, replace=False)
        x, y = x[indices], y[indices]

    return x, y


def get_fashion_mnist(n_samples=None):
    def load_mnist(path, kind='train'):
        """Load MNIST data from `path`"""
        labels_path = os.path.join(path, '%s-labels-idx1-ubyte.gz' % kind)
        images_path = os.path.join(path, '%s-images-idx3-ubyte.gz' % kind)

        with gzip.open(labels_path, 'rb') as lbpath:
            labels = np.frombuffer(lbpath.read(), dtype=np.uint8, offset=8)

        with gzip.open(images_path, 'rb') as imgpath:
            images = np.frombuffer(imgpath.read(), dtype=np.uint8,
                                   offset=16).reshape(len(labels), 784)

        return images, labels

    x_train, y_train = load_mnist(join(DATA_DIR, 'fmnist'), kind='train')
    x_test, y_test = load_mnist(join(DATA_DIR, 'fmnist'), kind='t10k')

    x = np.vstack((x_train, x_test))
    y = np.hstack((y_train, y_test))

    if n_samples is not None:
        indices = np.random.choice(list(range(x.shape[0])), n_samples, replace=False)
        x, y = x[indices], y[indices]

    return x, y


def get_mouse_60k(n_samples=None):
    with open(join(DATA_DIR, 'sc-mouse-60k-1k.pkl'), 'rb') as f:
        x = pickle.load(f)
    x = x.astype(np.float32).toarray()
    import json
    with open(join(DATA_DIR, 'sc-mouse-60k-1k-y.txt'), 'rb') as f:
        y = np.array(json.load(f), dtype=int)

    if n_samples is not None:
        indices = np.random.choice(list(range(x.shape[0])), n_samples, replace=False)
        x, y = x[indices], y[indices]

    return x, y


def check_error_approx():
    x, y = get_mouse_60k(1500)

    tsne = TSNE(
        perplexity=20, learning_rate=100, early_exaggeration=12,
        n_jobs=4, theta=0.5, initialization='pca', metric='euclidean',
        n_components=2, n_iter=750, early_exaggeration_iter=250,
        neighbors='exact', negative_gradient_method='bh',
        min_num_intervals=10, ints_in_interval=2,
        late_exaggeration_iter=0, late_exaggeration=4, callbacks=ErrorLogger(),
    )
    embedding = tsne.prepare_initial(x, initialization='random')

    errors = ErrorApproximations(embedding.affinities.P)
    logger = ErrorLogger()
    embedding.optimize(
        250, exaggeration=12, callbacks=[errors, logger],
        callbacks_every_iters=5, inplace=True,
    )
    embedding.optimize(
        750, exaggeration=None, callbacks=[errors, logger],
        callbacks_every_iters=5, inplace=True,
    )
    errors.report()

    plot(embedding, y)

    x = list(range(len(errors.exact_errors)))
    plt.semilogy(x, errors.exact_errors, label='Exact')
    plt.semilogy(x, errors.bh_errors, label='BH')
    plt.semilogy(x, errors.fft_errors, label='FFT')
    plt.legend()
    plt.show()


def run(perplexity=30, learning_rate=100, n_jobs=4):
    x, y = get_mouse_60k()
    # x, y = get_fashion_mnist()

    angle = 0.5
    ee = 12
    metric = 'euclidean'

    print(x.shape)

    start = time.time()
    tsne = TSNE(
        perplexity=perplexity, learning_rate=learning_rate, early_exaggeration=ee,
        n_jobs=n_jobs, theta=angle, initialization='random', metric=metric,
        n_components=2, n_iter=750, early_exaggeration_iter=250,
        neighbors='approx', negative_gradient_method='fft',
        min_num_intervals=10, ints_in_interval=1,
        late_exaggeration_iter=0, late_exaggeration=2., callbacks=ErrorLogger(),
    )
    # x = PCA(n_components=50).fit_transform(x)
    embedding = tsne.fit(x)
    print('-' * 80)
    print('tsne', time.time() - start)
    plt.title('tsne')
    plot(embedding, y)
    return

    x = np.ascontiguousarray(x.astype(np.float64))
    from fitsne import FItSNE
    start = time.time()
    embedding = FItSNE(
        x, 2, perplexity=perplexity, stop_lying_iter=250, ann_not_vptree=True,
        early_exag_coeff=ee, nthreads=n_jobs, theta=angle,
    )
    print('-' * 80)
    print('fft interp %.4f' % (time.time() - start))
    plt.title('fft interp')
    plot(embedding, y)
    plt.show()
    return

    init = PCA(n_components=2).fit_transform(x)
    start = time.time()
    embedding = MulticoreTSNE(
        early_exaggeration=ee, learning_rate=learning_rate, perplexity=perplexity,
        n_jobs=n_jobs, cheat_metric=False, angle=angle, init=init,
        metric=metric, verbose=True
    ).fit_transform(x)
    print('-' * 80)
    print('mctsne', time.time() - start)
    plt.title('mctsne')
    plot(embedding, y)
    plt.show()

    start = time.time()
    embedding = SKLTSNE(
        early_exaggeration=ee, learning_rate=learning_rate, angle=angle,
        perplexity=perplexity, init='pca', metric=metric,
    ).fit_transform(x)
    print('-' * 80)
    print('sklearn', time.time() - start)
    plt.title('sklearn')
    plot(embedding, y)
    plt.show()


def transform(n_jobs=4, grad='bh', neighbors='approx'):
    # iris = datasets.load_iris()
    # x, y = iris['data'], iris['target']
    x, y = get_mnist(20000)

    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.33, random_state=42)

    tsne = TSNE(
        n_components=2, perplexity=30, learning_rate=100, early_exaggeration=12,
        n_jobs=n_jobs, theta=0.5, initialization='random', metric='euclidean',
        n_iter=750, early_exaggeration_iter=250, neighbors=neighbors,
        negative_gradient_method=grad, min_num_intervals=10, ints_in_interval=2,
        late_exaggeration_iter=0, late_exaggeration=4, callbacks=[ErrorLogger()],
    )
    start = time.time()
    embedding = tsne.fit(x_train)
    print('tsne train', time.time() - start)

    plt.subplot(121)
    plot(embedding, y_train, show=False, ms=3)

    start = time.time()
    partial_embedding = embedding.transform(x_test, perplexity=20)
    # partial_embedding = embedding.get_partial_embedding_for(
    #     x_test, perplexity=10, initialization='random')
    # partial_embedding.optimize(200, exaggeration=2, inplace=True, momentum=0.1)
    print('tsne transform', time.time() - start)

    plt.subplot(122)
    plot(embedding, y_train, show=False, ms=3, alpha=0.25)
    plt.gca().set_color_cycle(None)
    plot(partial_embedding, y_test, show=False, ms=3, alpha=0.8)

    plt.show()


if __name__ == '__main__':
    fire.Fire()
